#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""VM names acquired from Nova API

Revision ID: 4f70135dc089
Revises: 4967af35820f
Create Date: 2019-01-08 14:18:11.909757

"""

# revision identifiers, used by Alembic.
revision = '4f70135dc089'
down_revision = '4967af35820f'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'apic_aim_vm_names',
        sa.Column('device_id', sa.String(36), nullable=False),
        sa.PrimaryKeyConstraint('device_id'),
        sa.Column('vm_name', sa.String(64), nullable=False),
    )
    op.create_table(
        'apic_aim_vm_name_updates',
        sa.Column('purpose', sa.String(36), nullable=False),
        sa.PrimaryKeyConstraint('purpose'),
        sa.Column('host_id', sa.String(36), nullable=False),
        sa.Column('last_incremental_update_time', sa.DateTime()),
        sa.Column('last_full_update_time', sa.DateTime()),
    )


def downgrade():
    pass



# Copyright (c) 2016 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

from unittest import mock

from neutron_lib import context
from neutron_lib.plugins import directory
from oslo_config import cfg

from gbpservice.neutron.tests.unit.plugins.ml2plus.drivers import (
    extension_test as ext_test)
from gbpservice.neutron.tests.unit.plugins.ml2plus import test_plugin


class ExtensionDriverTestCase(test_plugin.Ml2PlusPluginV2TestCase):

    _extension_drivers = ['test_ml2plus']

    def setUp(self):
        cfg.CONF.set_override('extension_drivers',
                self._extension_drivers, group='ml2')
        super(ExtensionDriverTestCase, self).setUp()
        self._plugin = directory.get_plugin()
        self._ctxt = context.get_admin_context()

    def _verify_subnetpool_create(self, code, exc_reason):
        tenant_id = str(uuid.uuid4())
        data = {'subnetpool': {'prefixes': ['10.0.0.0/8'],
                               'name': 'sp1',
                               'tenant_id': tenant_id}}
        req = self.new_create_request('subnetpools', data)
        res = req.get_response(self.api)
        self.assertEqual(code, res.status_int)

        subnetpool = self.deserialize(self.fmt, res)
        if exc_reason:
            self.assertEqual(exc_reason,
                             subnetpool['NeutronError']['type'])

        return (subnetpool, tenant_id)

    def _verify_subnetpool_update(self, subnetpool, code, exc_reason):
        sp_id = subnetpool['subnetpool']['id']
        new_name = 'a_brand_new_name'
        data = {'subnetpool': {'name': new_name}}
        req = self.new_update_request('subnetpools', data, sp_id)
        res = req.get_response(self.api)
        self.assertEqual(code, res.status_int)
        error = self.deserialize(self.fmt, res)
        self.assertEqual(exc_reason,
                         error['NeutronError']['type'])

    def test_faulty_process_create_subnetpool(self):
        with mock.patch.object(ext_test.TestExtensionDriver,
                               'process_create_subnetpool',
                               side_effect=TypeError):
            subnetpool, tenant_id = self._verify_subnetpool_create(
                500, 'HTTPInternalServerError')
            # Verify the operation is rolled back
            query_params = "tenant_id=%s" % tenant_id
            subnetpools = self._list('subnetpools', query_params=query_params)
            self.assertFalse(subnetpools['subnetpools'])

    def test_faulty_process_update_subnetpool(self):
        with mock.patch.object(ext_test.TestExtensionDriver,
                               'process_update_subnetpool',
                               side_effect=TypeError):
            subnetpool, tid = self._verify_subnetpool_create(201, None)
            self._verify_subnetpool_update(subnetpool, 500,
                                           'HTTPInternalServerError')

    def test_faulty_extend_subnetpool_dict(self):
        with mock.patch.object(ext_test.TestExtensionDriver,
                               'extend_subnetpool_dict',
                               side_effect=[None, None, TypeError]):
            subnetpool, tid = self._verify_subnetpool_create(201, None)
            self._verify_subnetpool_update(subnetpool, 400,
                                           'ExtensionDriverError')

    def test_subnetpool_attr(self):
        with self.subnetpool(['10.0.0.0/8'], name='sp1',
                             tenant_id='t1') as subnetpool:
            # Test create subnetpool
            ent = subnetpool['subnetpool'].get('subnetpool_extension')
            self.assertIsNotNone(ent)

            # Test list subnetpools
            res = self._list('subnetpools')
            val = res['subnetpools'][0].get('subnetpool_extension')
            self.assertEqual('Test_SubnetPool_Extension_extend', val)

            # Test subnetpool update
            data = {'subnetpool':
                    {'subnetpool_extension':
                     'Test_SubnetPool_Extension_Update'}}
            res = self._update('subnetpools', subnetpool['subnetpool']['id'],
                               data)
            val = res['subnetpool'].get('subnetpool_extension')
            self.assertEqual('Test_SubnetPool_Extension_Update_update', val)

    def test_extend_subnetpool_dict(self):
        with mock.patch.object(
                ext_test.TestExtensionDriver,
                'process_update_subnetpool') as pus, mock.patch.object(
                    ext_test.TestExtensionDriver,
                    'extend_subnetpool_dict') as esd, self.subnetpool(
                        ['10.0.0.0/8'], name='sp1',
                        tenant_id='t1') as subnetpool:
            subnetpool_id = subnetpool['subnetpool']['id']
            subnetpool_data = {'subnetpool': {'id': subnetpool_id}}
            self._plugin.update_subnetpool(self._ctxt, subnetpool_id,
                                           subnetpool_data)
            self.assertTrue(pus.called)
            self.assertTrue(esd.called)

    def _verify_address_scope_create(self, code, exc_reason):
        tenant_id = str(uuid.uuid4())
        data = {'address_scope': {'ip_version': 4,
                                  'name': 'as1',
                                  'tenant_id': tenant_id}}
        req = self.new_create_request('address-scopes', data)
        res = req.get_response(self.ext_api)
        self.assertEqual(code, res.status_int)

        address_scope = self.deserialize(self.fmt, res)
        if exc_reason:
            self.assertEqual(exc_reason,
                             address_scope['NeutronError']['type'])

        return (address_scope, tenant_id)

    def _verify_address_scope_update(self, address_scope, code, exc_reason):
        as_id = address_scope['address_scope']['id']
        new_name = 'a_brand_new_name'
        data = {'address_scope': {'name': new_name}}
        req = self.new_update_request('address-scopes', data, as_id)
        res = req.get_response(self.ext_api)
        self.assertEqual(code, res.status_int)
        error = self.deserialize(self.fmt, res)
        self.assertEqual(exc_reason,
                         error['NeutronError']['type'])

    def test_faulty_process_create_address_scope(self):
        with mock.patch.object(ext_test.TestExtensionDriver,
                               'process_create_address_scope',
                               side_effect=TypeError):
            address_scope, tenant_id = self._verify_address_scope_create(
                500, 'HTTPInternalServerError')
            # Verify the operation is rolled back
            query_params = "project_id=%s" % tenant_id
            address_scopes = self._list('address-scopes',
                                        query_params=query_params)
            self.assertFalse(address_scopes['address_scopes'])

    def test_faulty_process_update_address_scope(self):
        with mock.patch.object(ext_test.TestExtensionDriver,
                               'process_update_address_scope',
                               side_effect=TypeError):
            address_scope, tid = self._verify_address_scope_create(201, None)
            self._verify_address_scope_update(address_scope, 500,
                                              'HTTPInternalServerError')

    def test_faulty_extend_address_scope_dict(self):
        with mock.patch.object(ext_test.TestExtensionDriver,
                               'extend_address_scope_dict',
                               side_effect=[None, None, TypeError]):
            address_scope, tid = self._verify_address_scope_create(201, None)
            self._verify_address_scope_update(address_scope, 400,
                                              'ExtensionDriverError')

    def test_address_scope_attr(self):
        with self.address_scope(4, name='as1',
                                tenant_id='t1') as address_scope:
            # Test create address_scope
            ent = address_scope['address_scope'].get('address_scope_extension')
            self.assertIsNotNone(ent)

            # Test list address_scopes
            res = self._list('address-scopes')
            val = res['address_scopes'][0].get('address_scope_extension')
            self.assertEqual('Test_AddressScope_Extension_extend', val)

            # Test address_scope update
            data = {'address_scope':
                    {'address_scope_extension':
                     'Test_AddressScope_Extension_Update'}}
            res = self._update('address-scopes',
                               address_scope['address_scope']['id'], data)
            val = res['address_scope'].get('address_scope_extension')
            self.assertEqual('Test_AddressScope_Extension_Update_update', val)

    def test_extend_address_scope_dict(self):
        with mock.patch.object(
                ext_test.TestExtensionDriver,
                'process_update_address_scope') as puas, mock.patch.object(
                    ext_test.TestExtensionDriver,
                    'extend_address_scope_dict') as easd, self.address_scope(
                        4, name='as1', tenant_id='t1') as address_scope:
            address_scope_id = address_scope['address_scope']['id']
            address_scope_data = {'address_scope': {'id': address_scope_id}}
            self._plugin.update_address_scope(self._ctxt, address_scope_id,
                                              address_scope_data)
            self.assertTrue(puas.called)
            self.assertTrue(easd.called)


class DBExtensionDriverTestCase(test_plugin.Ml2PlusPluginV2TestCase):
    _extension_drivers = ['testdb_ml2plus']

    def setUp(self):
        cfg.CONF.set_override('extension_drivers',
                self._extension_drivers, group='ml2')
        super(DBExtensionDriverTestCase, self).setUp()
        self._plugin = directory.get_plugin()
        self._ctxt = context.get_admin_context()

    def test_subnetpool_attr(self):
        with self.subnetpool(['10.0.0.0/8'], name='sp1',
                             tenant_id='t1') as subnetpool:
            # Test create with default value.
            sp_id = subnetpool['subnetpool']['id']
            val = subnetpool['subnetpool']['subnetpool_extension']
            self.assertEqual("", val)
            res = self._show('subnetpools', sp_id)
            val = res['subnetpool']['subnetpool_extension']
            self.assertEqual("", val)

            # Test list.
            res = self._list('subnetpools')
            val = res['subnetpools'][0]['subnetpool_extension']
            self.assertEqual("", val)

        # Test create with explicit value.
        data = {'subnetpool':
                {'prefixes': ['10.0.0.0/8'],
                 'name': 'sp2',
                 'tenant_id': 't1',
                 'subnetpool_extension': 'abc'}}
        req = self.new_create_request('subnetpools', data, self.fmt)
        res = req.get_response(self.api)
        subnetpool = self.deserialize(self.fmt, res)
        subnetpool_id = subnetpool['subnetpool']['id']
        val = subnetpool['subnetpool']['subnetpool_extension']
        self.assertEqual("abc", val)
        res = self._show('subnetpools', subnetpool_id)
        val = res['subnetpool']['subnetpool_extension']
        self.assertEqual("abc", val)

        # Test update.
        data = {'subnetpool': {'subnetpool_extension': "def"}}
        res = self._update('subnetpools', subnetpool_id, data)
        val = res['subnetpool']['subnetpool_extension']
        self.assertEqual("def", val)
        res = self._show('subnetpools', subnetpool_id)
        val = res['subnetpool']['subnetpool_extension']
        self.assertEqual("def", val)

    def test_address_scope_attr(self):
        with self.address_scope(4, name='as1',
                                tenant_id='t1') as address_scope:
            # Test create with default value.
            as_id = address_scope['address_scope']['id']
            val = address_scope['address_scope']['address_scope_extension']
            self.assertEqual("", val)
            res = self._show('address-scopes', as_id)
            val = res['address_scope']['address_scope_extension']
            self.assertEqual("", val)

            # Test list.
            res = self._list('address-scopes')
            val = res['address_scopes'][0]['address_scope_extension']
            self.assertEqual("", val)

        # Test create with explicit value.
        data = {'address_scope':
                {'ip_version': 4,
                 'name': 'as2',
                 'tenant_id': 't1',
                 'address_scope_extension': 'abc'}}
        req = self.new_create_request('address-scopes', data, self.fmt)
        res = req.get_response(self.ext_api)
        address_scope = self.deserialize(self.fmt, res)
        address_scope_id = address_scope['address_scope']['id']
        val = address_scope['address_scope']['address_scope_extension']
        self.assertEqual("abc", val)
        res = self._show('address-scopes', address_scope_id)
        val = res['address_scope']['address_scope_extension']
        self.assertEqual("abc", val)

        # Test update.
        data = {'address_scope': {'address_scope_extension': "def"}}
        res = self._update('address-scopes', address_scope_id, data)
        val = res['address_scope']['address_scope_extension']
        self.assertEqual("def", val)
        res = self._show('address-scopes', address_scope_id)
        val = res['address_scope']['address_scope_extension']
        self.assertEqual("def", val)



# Copyright 2014 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""gbp_rule_servicechain_mapping

Revision ID: ceba6e091b2a
Revises: 577bb4469944
Create Date: 2014-10-25 17:43:08.98888

"""

# revision identifiers, used by Alembic.
revision = 'ceba6e091b2a'
down_revision = '577bb4469944'

from alembic import op
import sqlalchemy as sa


def upgrade(active_plugins=None, options=None):
    op.create_table(
        'gpm_service_policy_ipaddress_mappings',
        sa.Column('service_policy_id', sa.String(length=36), nullable=False),
        sa.Column('ipaddress', sa.String(length=36)),
        sa.Column('policy_target_group', sa.String(length=36)),
        sa.PrimaryKeyConstraint('policy_target_group', 'service_policy_id'),
        sa.ForeignKeyConstraint(['policy_target_group'],
                                ['gp_policy_target_groups.id'],
                                ondelete='CASCADE'),
        sa.ForeignKeyConstraint(['service_policy_id'],
                                ['gp_network_service_policies.id'],
                                ondelete='CASCADE'),
    )


def downgrade(active_plugins=None, options=None):
    op.drop_table('gpm_service_policy_ipaddress_mappings')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.db import model_base
from oslo_log import log as logging
import sqlalchemy as sa

from gbpservice.neutron.db.grouppolicy import group_policy_db as gp_db

LOG = logging.getLogger(__name__)
PROVIDER = 'provider'
CONSUMER = 'consumer'
MANAGEMENT = 'management'
RELATIONSHIPS = [PROVIDER, CONSUMER, MANAGEMENT]


class NodeToDriverMapping(model_base.BASEV2):
    """Node to Driver mapping DB.

    This table keeps track of the driver owning a specific SC Node based on
    the SC instance
    """

    __tablename__ = 'ncp_node_to_driver_mapping'
    servicechain_node_id = sa.Column(sa.String(36),
                                     sa.ForeignKey('sc_nodes.id',
                                                   ondelete='CASCADE'),
                                     nullable=False, primary_key=True)
    # Based on the extension name
    driver_name = sa.Column(sa.String(36), nullable=False)
    servicechain_instance_id = sa.Column(sa.String(36),
                                         sa.ForeignKey('sc_instances.id',
                                                       ondelete='CASCADE'),
                                         primary_key=True)


class ServiceTarget(model_base.BASEV2):
    """Service related policy targets.

    Internal information regarding the policy targets owned by services.
    """

    __tablename__ = 'ncp_service_targets'
    policy_target_id = sa.Column(sa.String(36),
                                 sa.ForeignKey(gp_db.PolicyTarget.id,
                                               ondelete='CASCADE'),
                                 nullable=False, primary_key=True)
    # Not a FK to avoid constraint error on SCI delete
    # keeping the DB entry is useful to identify uncleaned PTs
    servicechain_instance_id = sa.Column(sa.String(36),
                                         nullable=False, primary_key=True)
    # Not a FK to avoid constraint error on SCN delete.
    # keeping the DB entry is useful to identify uncleaned PTs
    servicechain_node_id = sa.Column(sa.String(36),
                                     nullable=False, primary_key=True)
    # Defines on which "side" of the chain the PT is placed. typically
    # its values can be "provider", "consumer" or "management"
    relationship = sa.Column(sa.String(25), nullable=False)
    position = sa.Column(sa.Integer)


def set_node_owner(context, driver_name):
    session = context.session
    with session.begin(subtransactions=True):
        owner = NodeToDriverMapping(
            servicechain_instance_id=context.instance['id'],
            servicechain_node_id=context.current_node['id'],
            driver_name=driver_name)
        session.add(owner)


def get_node_owner(context):
    session = context.session
    with session.begin(subtransactions=True):
        query = session.query(NodeToDriverMapping)
        query = query.filter_by(
            servicechain_instance_id=context.instance['id'])
        query = query.filter_by(
            servicechain_node_id=context.current_node['id'])
        return query.all()


def unset_node_owner(context):
    session = context.session
    with session.begin(subtransactions=True):
        query = session.query(NodeToDriverMapping)
        query = query.filter_by(
            servicechain_instance_id=context.instance['id'])
        query = query.filter_by(
            servicechain_node_id=context.current_node['id'])
        for owner in query.all():
            session.delete(owner)


def set_service_target(context, policy_target_id, relationship):
    session = context.session
    with session.begin(subtransactions=True):
        owner = ServiceTarget(
            policy_target_id=policy_target_id,
            servicechain_instance_id=context.instance['id'],
            servicechain_node_id=context.current_node['id'],
            position=context.current_position,
            relationship=relationship)
        session.add(owner)


def get_service_targets(session, policy_target_id=None, relationship=None,
                        servicechain_instance_id=None, position=None,
                        servicechain_node_id=None):
    with session.begin(subtransactions=True):
        query = _prepare_service_target_query(
            session, policy_target_id=policy_target_id,
            relationship=relationship,
            servicechain_instance_id=servicechain_instance_id,
            position=position, servicechain_node_id=servicechain_node_id)
        return query.all()


def get_service_targets_count(session, policy_target_id=None,
                              relationship=None, servicechain_instance_id=None,
                              position=None, servicechain_node_id=None):
    with session.begin(subtransactions=True):
        query = _prepare_service_target_query(
            session, policy_target_id=policy_target_id,
            relationship=relationship,
            servicechain_instance_id=servicechain_instance_id,
            position=position, servicechain_node_id=servicechain_node_id)
        return query.count()


def _prepare_service_target_query(session, policy_target_id=None,
                                  relationship=None,
                                  servicechain_instance_id=None, position=None,
                                  servicechain_node_id=None):
    query = session.query(ServiceTarget)
    if servicechain_instance_id:
        query = query.filter_by(
            servicechain_instance_id=servicechain_instance_id)
    if servicechain_node_id:
        query = query.filter_by(
            servicechain_node_id=servicechain_node_id)
    if policy_target_id:
        query = query.filter_by(policy_target_id=policy_target_id)
    if position:
        query = query.filter_by(position=position)
    if relationship:
        query = query.filter_by(relationship=relationship)
    return query



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Fix missing nested domain DB migration

Revision ID: 27583c259fa7
Revises: 799f0516bc08
Create Date: 2020-05-27 14:18:11.909757

"""

# revision identifiers, used by Alembic.
revision = '27583c259fa7'
down_revision = '799f0516bc08'

import os
import sys

from neutron.db import migration

from oslo_utils import importutils

from gbpservice.neutron.db.migration import alembic_migrations as am

# This is a hack to get around the fact that the versions
# directory has no __init__.py
filepath = os.path.abspath(am.__file__)
basepath = filepath[:filepath.rfind("/")] + "/versions"
sys.path.append(basepath)

DB_4967af35820f = '4967af35820f_cisco_apic_nested_domain'


def ensure_4967af35820f_migration():
    if not migration.schema_has_table(
            'apic_aim_network_nested_domain_allowed_vlans'):
        db_4967af35820f = importutils.import_module(DB_4967af35820f)
        db_4967af35820f.upgrade()


def upgrade():
    ensure_4967af35820f_migration()
    # remove the appended path
    del sys.path[sys.path.index(basepath)]


def downgrade():
    pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""application_policy_group

Revision ID: b6e301d5757f
Revises: daaa11a358a2
Create Date: 2017-02-10 01:15:32.361753

"""

# revision identifiers, used by Alembic.
revision = 'b6e301d5757f'
down_revision = 'daaa11a358a2'

from alembic import op
import sqlalchemy as sa


def upgrade():

    op.create_table(
        'gp_application_policy_groups',
        sa.Column('id', sa.String(36), nullable=False),
        sa.Column('tenant_id', sa.String(length=255), nullable=True),
        sa.Column('name', sa.String(length=50), nullable=True),
        sa.Column('description', sa.String(length=255), nullable=True),
        sa.Column('status', sa.String(length=16), nullable=True),
        sa.Column('status_details', sa.String(length=4096), nullable=True),
        sa.PrimaryKeyConstraint('id'))

    op.add_column(
        'gp_policy_target_groups',
        sa.Column('application_policy_group_id', sa.String(length=36),
                  nullable=True))

    op.create_foreign_key('gp_application_policy_group_ibfk_1',
                          source_table='gp_policy_target_groups',
                          referent_table='gp_application_policy_groups',
                          local_cols=['application_policy_group_id'],
                          remote_cols=['id'])


def downgrade():
    pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from alembic import config as alembic_config
from neutron.db.migration.cli import CONF
from oslo_config import cfg


global_opts = [
    cfg.StrOpt('apic_system_id',
               help="Prefix for APIC domain/names/profiles created"),
]
CONF.register_cli_opts(global_opts)


def main():
    config = alembic_config.Config(
        os.path.join(os.path.dirname(__file__), 'alembic.ini'))
    config.set_main_option(
        'script_location',
        'gbpservice.neutron.db.migration:alembic_migrations')
    config.neutron_config = CONF
    CONF(project='neutron')
    CONF.command.func(config, CONF.command.name)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from unittest import mock

from neutron.tests.unit.db import test_db_base_plugin_v2
from neutron_lib import context
from neutronclient.common import exceptions
from neutronclient.v2_0 import client
from oslo_config import cfg

from gbpservice.network.neutronv2 import client as neutronclient


CONF = cfg.CONF

# NOTE: Neutron client raises Exception which is discouraged by HACKING.
#       We set this variable here and use it for assertions below to avoid
#       the hacking checks until we can make neutron client throw a custom
#       exception class instead.
NEUTRON_CLIENT_EXCEPTION = Exception


class TestNeutronClient(test_db_base_plugin_v2.NeutronDbPluginV2TestCase):

    def setUp(self):
        super(TestNeutronClient, self).setUp()

    def test_withtoken(self):
        CONF.set_override('neutron_server_url',
                          'http://anyhost',
                          group='neutron')
        CONF.set_override('url_timeout',
                          30,
                          group='neutron')
        my_context = context.ContextBase('userid',
                                         'my_tenantid',
                                         auth_token='token')
        cl = neutronclient.get_client(my_context)

        self.assertEqual(CONF.neutron.neutron_server_url,
                         cl.httpclient.endpoint_url)
        self.assertEqual(my_context.auth_token,
                         cl.httpclient.auth_token)
        self.assertEqual(CONF.neutron.url_timeout, cl.httpclient.timeout)

    def test_withouttoken(self):
        my_context = context.ContextBase('userid', 'my_tenantid')
        self.assertRaises(exceptions.Unauthorized,
                          neutronclient.get_client,
                          my_context)

    def test_withtoken_context_is_admin(self):
        CONF.set_override('neutron_server_url',
                          'http://anyhost',
                          group='neutron')
        CONF.set_override('url_timeout',
                          30,
                          group='neutron')
        my_context = context.ContextBase('userid',
                                         'my_tenantid',
                                         auth_token='token',
                                         is_admin=True)
        cl = neutronclient.get_client(my_context)

        self.assertEqual(CONF.neutron.neutron_server_url,
                         cl.httpclient.endpoint_url)
        self.assertEqual(my_context.auth_token,
                         cl.httpclient.auth_token)
        self.assertEqual(CONF.neutron.url_timeout, cl.httpclient.timeout)

    def test_withouttoken_keystone_connection_error(self):
        CONF.set_override('neutron_server_url',
                          'http://anyhost',
                          group='neutron')
        CONF.set_override('auth_strategy',
                          'keystone',
                          group='neutron')
        my_context = context.ContextBase('userid', 'my_tenantid')
        self.assertRaises(NEUTRON_CLIENT_EXCEPTION,
                          neutronclient.get_client,
                          my_context)

    def test_reuse_admin_token(self):
        CONF.set_override('neutron_server_url',
                          'http://anyhost',
                          group='neutron')
        CONF.set_override('url_timeout',
                          30,
                          group='neutron')
        token_store = neutronclient.AdminTokenStore.get()
        token_store.admin_auth_token = 'new_token'
        my_context = context.ContextBase('userid', 'my_tenantid',
                                         auth_token='token')
        with mock.patch.object(client.Client, "list_networks",
                               side_effect=mock.Mock):
            with mock.patch.object(client.Client, 'get_auth_info',
                                   return_value={'auth_token': 'new_token1'}):
                client1 = neutronclient.get_client(my_context, True)
                client1.list_networks(retrieve_all=False)
                self.assertEqual('new_token1', token_store.admin_auth_token)
                client1 = neutronclient.get_client(my_context, True)
                client1.list_networks(retrieve_all=False)
                self.assertEqual('new_token1', token_store.admin_auth_token)

    def test_admin_token_updated(self):
        CONF.set_override('neutron_server_url',
                          'http://anyhost',
                          group='neutron')
        CONF.set_override('url_timeout',
                          30,
                          group='neutron')
        token_store = neutronclient.AdminTokenStore.get()
        token_store.admin_auth_token = 'new_token'
        tokens = [{'auth_token': 'new_token1'}, {'auth_token': 'new_token'}]
        my_context = context.ContextBase('userid', 'my_tenantid',
                                         auth_token='token')
        with mock.patch.object(client.Client, "list_networks",
                               side_effect=mock.Mock):
            with mock.patch.object(client.Client, 'get_auth_info',
                                   side_effect=tokens.pop):
                client1 = neutronclient.get_client(my_context, True)
                client1.list_networks(retrieve_all=False)
                self.assertEqual('new_token', token_store.admin_auth_token)
                client1 = neutronclient.get_client(my_context, True)
                client1.list_networks(retrieve_all=False)
                self.assertEqual('new_token1', token_store.admin_auth_token)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.db import models_v2
from neutron_lib.db import model_base
from neutron_lib import exceptions as nexc
from neutron_lib.plugins import directory
from oslo_log import helpers as log
from oslo_utils import uuidutils
import sqlalchemy as sa
from sqlalchemy import orm

from gbpservice._i18n import _
from gbpservice.common import utils as gbp_utils
from gbpservice.neutron.db import api as db_api
from gbpservice.neutron.db.grouppolicy import group_policy_db as gpdb
from gbpservice.neutron.extensions import group_policy as gpolicy
from gbpservice.neutron.services.grouppolicy.common import exceptions
from gbpservice.neutron.services.grouppolicy.common import utils


def get_current_context():
    return gbp_utils.get_current_context()


class AddressScopeUpdateForL3PNotSupported(nexc.BadRequest):
    message = _("Address Scope update for L3 Policy is not supported.")


class PolicyTargetMapping(gpdb.PolicyTarget):
    """Mapping of PolicyTarget to Neutron Port."""
    __table_args__ = {'extend_existing': True}
    __mapper_args__ = {'polymorphic_identity': 'mapping'}
    # REVISIT(ivar): Set null on delete is a temporary workaround until Nova
    # bug 1158684 is fixed.
    port_id = sa.Column(sa.String(36), sa.ForeignKey('ports.id',
                                                     ondelete='SET NULL'),
                        nullable=True, unique=True)


class PTGToSubnetAssociation(model_base.BASEV2):
    """Many to many relation between PolicyTargetGroup and Subnets."""
    __tablename__ = 'gp_ptg_to_subnet_associations'
    policy_target_group_id = sa.Column(
        sa.String(36), sa.ForeignKey('gp_policy_target_groups.id'),
        primary_key=True)
    subnet_id = sa.Column(sa.String(36), sa.ForeignKey('subnets.id'),
                          primary_key=True)


class PolicyTargetGroupMapping(gpdb.PolicyTargetGroup):
    """Mapping of PolicyTargetGroup to set of Neutron Subnets."""
    __table_args__ = {'extend_existing': True}
    __mapper_args__ = {'polymorphic_identity': 'mapping'}
    subnets = orm.relationship(PTGToSubnetAssociation,
                               cascade='all', lazy="joined")


class L2PolicyMapping(gpdb.L2Policy):
    """Mapping of L2Policy to Neutron Network."""
    __table_args__ = {'extend_existing': True}
    __mapper_args__ = {'polymorphic_identity': 'mapping'}
    network_id = sa.Column(sa.String(36), sa.ForeignKey('networks.id'),
                           nullable=True, unique=True)


class L3PolicyRouterAssociation(model_base.BASEV2):
    """Models the many to many relation between L3Policies and Routers."""
    __tablename__ = 'gp_l3_policy_router_associations'
    l3_policy_id = sa.Column(sa.String(36), sa.ForeignKey('gp_l3_policies.id'),
                             primary_key=True)
    router_id = sa.Column(sa.String(36), sa.ForeignKey('routers.id'),
                          primary_key=True)


class L3PolicySubnetpoolV4Association(model_base.BASEV2):
    """Models one to many relation between a L3Policy and v4 Subnetpools."""
    __tablename__ = 'gp_l3_policy_subnetpool_v4_associations'
    l3_policy_id = sa.Column(sa.String(36), sa.ForeignKey('gp_l3_policies.id'),
                             primary_key=True)
    subnetpool_id = sa.Column(
        sa.String(36), sa.ForeignKey('subnetpools.id'), primary_key=True)


class L3PolicySubnetpoolV6Association(model_base.BASEV2):
    """Models one to many relation between a L3Policy and v6 Subnetpools."""
    __tablename__ = 'gp_l3_policy_subnetpool_v6_associations'
    l3_policy_id = sa.Column(sa.String(36), sa.ForeignKey('gp_l3_policies.id'),
                             primary_key=True)
    subnetpool_id = sa.Column(
        sa.String(36), sa.ForeignKey('subnetpools.id'), primary_key=True)


class L3PolicyMapping(gpdb.L3Policy):
    """Mapping of L3Policy to set of Neutron Resources."""
    __table_args__ = {'extend_existing': True}
    __mapper_args__ = {'polymorphic_identity': 'mapping'}
    routers = orm.relationship(L3PolicyRouterAssociation,
                               cascade='all', lazy="joined")
    address_scope_v4_id = sa.Column(
        sa.String(36), sa.ForeignKey('address_scopes.id'),
        nullable=True, unique=True)
    address_scope_v6_id = sa.Column(
        sa.String(36), sa.ForeignKey('address_scopes.id'),
        nullable=True, unique=True)
    subnetpools_v4 = orm.relationship(L3PolicySubnetpoolV4Association,
                                      cascade='all', lazy="joined")
    subnetpools_v6 = orm.relationship(L3PolicySubnetpoolV6Association,
                                      cascade='all', lazy="joined")


class ExternalSegmentMapping(gpdb.ExternalSegment):
    """Mapping of L2Policy to Neutron Network."""
    __table_args__ = {'extend_existing': True}
    __mapper_args__ = {'polymorphic_identity': 'mapping'}
    subnet_id = sa.Column(sa.String(36), sa.ForeignKey('subnets.id'),
                          nullable=True, unique=True)


class NATPoolMapping(gpdb.NATPool):
    """Mapping of NAT Pool to Neutron Subnet."""
    __table_args__ = {'extend_existing': True}
    __mapper_args__ = {'polymorphic_identity': 'mapping'}
    subnet_id = sa.Column(sa.String(36), sa.ForeignKey('subnets.id'),
                          nullable=True, unique=True)


class GroupPolicyMappingDbPlugin(gpdb.GroupPolicyDbPlugin):
    """Group Policy Mapping interface implementation using SQLAlchemy models.
    """

    def _make_policy_target_dict(self, pt, fields=None, **kwargs):
        res = super(GroupPolicyMappingDbPlugin,
                    self)._make_policy_target_dict(pt)
        res['port_id'] = pt.port_id
        res.update(kwargs)
        return db_api.resource_fields(res, fields)

    def _make_policy_target_group_dict(self, ptg, fields=None):
        res = super(GroupPolicyMappingDbPlugin,
                    self)._make_policy_target_group_dict(ptg)
        res['subnets'] = [subnet.subnet_id for subnet in ptg.subnets]
        return db_api.resource_fields(res, fields)

    def _get_subnetpools(self, id_list):
        context = get_current_context().elevated()
        with context.session.begin(subtransactions=True):
            filters = {'id': id_list}
            return db_api.get_collection_query(
                context, models_v2.SubnetPool, filters=filters).all()

    def _make_l2_policy_dict(self, l2p, fields=None):
        res = super(GroupPolicyMappingDbPlugin,
                    self)._make_l2_policy_dict(l2p)
        res['network_id'] = l2p.network_id
        return db_api.resource_fields(res, fields)

    def _make_l3_policy_dict(self, l3p, fields=None, ip_pool=None):
        res = super(GroupPolicyMappingDbPlugin,
                    self)._make_l3_policy_dict(l3p)
        res['routers'] = [router.router_id for router in l3p.routers]
        res['address_scope_v4_id'] = l3p.address_scope_v4_id
        res['address_scope_v6_id'] = l3p.address_scope_v6_id
        res['subnetpools_v4'] = [sp.subnetpool_id for sp in l3p.subnetpools_v4]
        res['subnetpools_v6'] = [sp.subnetpool_id for sp in l3p.subnetpools_v6]
        if ip_pool:
            res['ip_pool'] = ip_pool
        subnetpools = self._get_subnetpools(res['subnetpools_v4'] +
                                            res['subnetpools_v6'])
        pool_list = [prefix['cidr'] for pool in subnetpools
                     for prefix in pool['prefixes']]
        if pool_list:
            res['ip_pool'] = utils.convert_ip_pool_list_to_string(
                pool_list)
        return db_api.resource_fields(res, fields)

    def _make_external_segment_dict(self, es, fields=None):
        res = super(GroupPolicyMappingDbPlugin,
                    self)._make_external_segment_dict(es)
        res['subnet_id'] = es.subnet_id
        return db_api.resource_fields(res, fields)

    def _make_nat_pool_dict(self, np, fields=None):
        res = super(GroupPolicyMappingDbPlugin,
                    self)._make_nat_pool_dict(np)
        res['subnet_id'] = np.subnet_id
        return db_api.resource_fields(res, fields)

    def _set_port_for_policy_target(self, context, pt_id, port_id):
        with context.session.begin(subtransactions=True):
            pt_db = self._get_policy_target(context, pt_id)
            pt_db.port_id = port_id

    def _add_subnet_to_policy_target_group(self, context, ptg_id, subnet_id):
        with context.session.begin(subtransactions=True):
            ptg_db = self._get_policy_target_group(context, ptg_id)
            assoc = PTGToSubnetAssociation(policy_target_group_id=ptg_id,
                                           subnet_id=subnet_id)
            ptg_db.subnets.append(assoc)
        return [subnet.subnet_id for subnet in ptg_db.subnets]

    def _remove_subnets_from_policy_target_groups(self, context, subnet_ids):
        with context.session.begin(subtransactions=True):
            assocs = context.session.query(PTGToSubnetAssociation).filter(
                PTGToSubnetAssociation.subnet_id.in_(subnet_ids)).all()
            for assoc in assocs:
                context.session.delete(assoc)

    def _remove_subnets_from_policy_target_group(self, context, ptg_id):
        with context.session.begin(subtransactions=True):
            ptg_db = self._get_policy_target_group(context, ptg_id)
            assocs = (context.session.query(PTGToSubnetAssociation).
                      filter_by(policy_target_group_id=ptg_id).all())
            ptg_db.update({'subnets': []})
            for assoc in assocs:
                context.session.delete(assoc)
        return []

    def _set_network_for_l2_policy(self, context, l2p_id, network_id):
        with context.session.begin(subtransactions=True):
            l2p_db = self._get_l2_policy(context, l2p_id)
            l2p_db.network_id = network_id

    def _set_address_scope_for_l3_policy_by_id(
        self, context, l3p_id, address_scope_id, ip_version=4):
        with context.session.begin(subtransactions=True):
            l3p_db = self._get_l3_policy(context, l3p_id)
            self._set_address_scope_for_l3_policy(
                context, l3p_db, address_scope_id, ip_version)

    def _set_address_scope_for_l3_policy(self, context, l3p_db,
                                         address_scope_id, ip_version=4):
        if not address_scope_id:
            return
        # TODO(Sumit): address_scope_id validation
        with context.session.begin(subtransactions=True):
            if ip_version == 4:
                l3p_db.address_scope_v4_id = address_scope_id
            else:
                l3p_db.address_scope_v6_id = address_scope_id

    def _add_subnetpool_to_l3_policy_by_id(self, context, l3p_id,
                                           subnetpool_id, ip_version=4):
        with context.session.begin(subtransactions=True):
            l3p_db = self._get_l3_policy(context, l3p_id)
            return self._add_subnetpool_to_l3_policy(
                context, l3p_db, subnetpool_id, ip_version)

    def _add_subnetpool_to_l3_policy(self, context, l3p_db,
                                     subnetpool_id, ip_version=4):
        # TODO(Sumit): subnetpool_id validation
        with context.session.begin(subtransactions=True):
            if ip_version == 4:
                assoc = L3PolicySubnetpoolV4Association(
                    l3_policy_id=l3p_db['id'], subnetpool_id=subnetpool_id)
                l3p_db.subnetpools_v4.append(assoc)
            else:
                assoc = L3PolicySubnetpoolV6Association(
                    l3_policy_id=l3p_db['id'], subnetpool_id=subnetpool_id)
                l3p_db.subnetpools_v6.append(assoc)
        return {4: [sp.subnetpool_id for sp in l3p_db.subnetpools_v4],
                6: [sp.subnetpool_id for sp in l3p_db.subnetpools_v6]}

    def _add_subnetpools_to_l3_policy_by_id(self, context, l3p_id,
                                            subnetpools, ip_version=4):
        with context.session.begin(subtransactions=True):
            l3p_db = self._get_l3_policy(context, l3p_id)
            self._add_subnetpools_to_l3_policy(
                context, l3p_db, subnetpools, ip_version)

    def _add_subnetpools_to_l3_policy(self, context, l3p_db,
                                      subnetpools, ip_version=4):
        for sp in subnetpools or []:
            self._add_subnetpool_to_l3_policy(
                context, l3p_db, sp, ip_version=ip_version)

    def _remove_subnetpool_from_l3_policy(self, context, l3p_id,
                                          subnetpool_id, ip_version=4):
        with context.session.begin(subtransactions=True):
            l3p_db = self._get_l3_policy(context, l3p_id)
            if ip_version == 4:
                assoc = (context.session.query(
                    L3PolicySubnetpoolV4Association).filter_by(
                        l3_policy_id=l3p_id,
                        subnetpool_id=subnetpool_id).one())
                l3p_db.subnetpools_v4.remove(assoc)
            else:
                assoc = (context.session.query(
                    L3PolicySubnetpoolV6Association).filter_by(
                        l3_policy_id=l3p_id,
                        subnetpool_id=subnetpool_id).one())
                l3p_db.subnetpools_v6.remove(assoc)
            context.session.delete(assoc)
        return {4: [sp.subnetpool_id for sp in l3p_db.subnetpools_v4],
                6: [sp.subnetpool_id for sp in l3p_db.subnetpools_v6]}

    def _update_subnetpools_for_l3_policy(self, context, l3p_id,
                                          subnetpools, ip_version=4):
        # Add/remove associations for changes in subnetpools
        # TODO(Sumit): Before disassociating a subnetpool, check that
        # there is no PT present on a subnet which belongs to that subnetpool
        if not subnetpools:
            return
        with context.session.begin(subtransactions=True):
            l3p_db = self._get_l3_policy(context, l3p_id)
            new_subnetpools = set(subnetpools)
            if ip_version == 4:
                old_subnetpools = set(sp.subnetpool_id
                                      for sp in l3p_db.subnetpools_v4)
            else:
                old_subnetpools = set(sp.subnetpool_id
                                      for sp in l3p_db.subnetpools_v6)
            for sp in new_subnetpools - old_subnetpools:
                if ip_version == 4:
                    assoc = L3PolicySubnetpoolV4Association(
                        l3_policy_id=l3p_id, subnetpool_id=sp)
                    l3p_db.subnetpools_v4.append(assoc)
                else:
                    assoc = L3PolicySubnetpoolV6Association(
                        l3_policy_id=l3p_id, subnetpool_id=sp)
                    l3p_db.subnetpools_v6.append(assoc)
            for sp in old_subnetpools - new_subnetpools:
                if ip_version == 4:
                    assoc = (context.session.query(
                        L3PolicySubnetpoolV4Association).filter_by(
                            l3_policy_id=l3p_id, subnetpool_id=sp).one())
                    l3p_db.subnetpools_v4.remove(assoc)
                else:
                    assoc = (context.session.query(
                        L3PolicySubnetpoolV6Association).filter_by(
                            l3_policy_id=l3p_id, subnetpool_id=sp).one())
                    l3p_db.subnetpools_v6.remove(assoc)
                context.session.delete(assoc)

    def _add_router_to_l3_policy(self, context, l3p_id, router_id):
        with context.session.begin(subtransactions=True):
            l3p_db = self._get_l3_policy(context, l3p_id)
            assoc = L3PolicyRouterAssociation(l3_policy_id=l3p_id,
                                              router_id=router_id)
            l3p_db.routers.append(assoc)
        return [router.router_id for router in l3p_db.routers]

    def _remove_router_from_l3_policy(self, context, l3p_id, router_id):
        with context.session.begin(subtransactions=True):
            l3p_db = self._get_l3_policy(context, l3p_id)
            assoc = (context.session.query(L3PolicyRouterAssociation).
                     filter_by(l3_policy_id=l3p_id, router_id=router_id).
                     one())
            l3p_db.routers.remove(assoc)
            context.session.delete(assoc)
        return [router.router_id for router in l3p_db.routers]

    def _set_subnet_to_es(self, context, es_id, subnet_id):
        with context.session.begin(subtransactions=True):
            es_db = self._get_external_segment(context, es_id)
            es_db.subnet_id = subnet_id

    def _set_subnet_to_np(self, context, np_id, subnet_id):
        with context.session.begin(subtransactions=True):
            np_db = self._get_nat_pool(context, np_id)
            np_db.subnet_id = subnet_id

    def _update_ess_for_l3p(self, context, l3p_id, ess):
        with context.session.begin(subtransactions=True):
            l3p_db = self._get_l3_policy(context, l3p_id)
            self._set_ess_for_l3p(context, l3p_db, ess)

    def _get_nat_pool(self, context, nat_pool_id):
        return self._find_gbp_resource(
            context, NATPoolMapping, nat_pool_id,
            gpolicy.NATPoolNotFound)

    def _get_l3p_ptgs(self, context, l3p_id):
        return super(GroupPolicyMappingDbPlugin, self)._get_l3p_ptgs(
            context, l3p_id, l3p_klass=L3PolicyMapping,
            ptg_klass=PolicyTargetGroupMapping, l2p_klass=L2PolicyMapping)

    def get_l3p_id_from_router_id(self, context, router_id):
        mapping = context.session.query(L3PolicyRouterAssociation).filter_by(
            router_id=router_id).first()
        if mapping:
            return mapping['l3_policy_id']

    def _set_db_np_subnet(self, context, nat_pool, subnet_id):
        with context.session.begin(subtransactions=True):
            nat_pool['subnet_id'] = subnet_id
            db_np = self._get_nat_pool(context, nat_pool['id'])
            db_np.subnet_id = nat_pool['subnet_id']
            context.session.merge(db_np)

    def _get_ptgs_for_subnet(self, context, subnet_id):
        return [x['policy_target_group_id'] for x in
            context.session.query(PTGToSubnetAssociation).filter_by(
                subnet_id=subnet_id)]

    def _validate_pt_port_exta_attributes(self, context, pt):
        attributes = pt.get('port_attributes')
        if attributes:
            # Check network ID not overridden
            if 'network_id' in attributes:
                raise exceptions.InvalidPortExtraAttributes(
                    attribute='network_id', reason='read only attribute')
            if 'fixed_ips' in attributes:
                ptg = self.get_policy_target_group(
                    context, pt['policy_target_group_id'])
                subnets = ptg['subnets']
                for fixed_ip in attributes.get('fixed_ips'):
                    if ('subnet_id' in fixed_ip and
                            fixed_ip['subnet_id'] not in subnets):
                        raise exceptions.InvalidPortExtraAttributes(
                            attribute='fixed_ips:subnet_id',
                            reason='subnet not in PTG')
            if 'allowed_address_pairs' in attributes:
                # REVISIT(ivar); Could be allowed with certain restrictions,
                # but we don't have a use case for it right now
                raise exceptions.InvalidPortExtraAttributes(
                    attribute='allowed_address_pairs',
                    reason='read only attribute')

    @log.log_method_call
    def create_policy_target(self, context, policy_target):
        pt = policy_target['policy_target']
        tenant_id = self._get_tenant_id_for_create(context, pt)
        with db_api.CONTEXT_WRITER.using(context):
            self._validate_pt_port_exta_attributes(context, pt)
            pt_db = PolicyTargetMapping(
                id=uuidutils.generate_uuid(),
                tenant_id=tenant_id,
                name=pt['name'],
                description=pt['description'],
                policy_target_group_id=pt['policy_target_group_id'],
                port_id=pt['port_id'],
                cluster_id=pt['cluster_id'])
            context.session.add(pt_db)
            return self._make_policy_target_dict(
                pt_db, port_attributes=pt.get('port_attributes', {}))

    @log.log_method_call
    def get_policy_targets_count(self, context, filters=None):
        return db_api.get_collection_count(context, PolicyTargetMapping,
                                           filters=filters)

    @log.log_method_call
    def get_policy_targets(self, context, filters=None, fields=None,
                           sorts=None, limit=None, marker=None,
                           page_reverse=False):
        plugin = directory.get_plugin()
        marker_obj = db_api.get_marker_obj(plugin, context,
                                           'policy_target', limit,
                                           marker)
        return db_api.get_collection(context, PolicyTargetMapping,
                                     self._make_policy_target_dict,
                                     filters=filters, fields=fields,
                                     sorts=sorts, limit=limit,
                                     marker_obj=marker_obj,
                                     page_reverse=page_reverse)

    @log.log_method_call
    def create_policy_target_group(self, context, policy_target_group):
        ptg = policy_target_group['policy_target_group']
        tenant_id = self._get_tenant_id_for_create(context, ptg)
        with db_api.CONTEXT_WRITER.using(context):
            if ptg['service_management']:
                self._validate_service_management_ptg(context, tenant_id)
            uuid = ptg.get('id')
            if not uuid:
                uuid = uuidutils.generate_uuid()
            ptg_db = PolicyTargetGroupMapping(
                id=uuid, tenant_id=tenant_id,
                name=ptg['name'], description=ptg['description'],
                l2_policy_id=ptg['l2_policy_id'],
                application_policy_group_id=ptg.get(
                    'application_policy_group_id', None),
                network_service_policy_id=ptg['network_service_policy_id'],
                shared=ptg.get('shared', False),
                service_management=ptg.get('service_management', False))
            context.session.add(ptg_db)
            if 'subnets' in ptg:
                for subnet in ptg['subnets']:
                    assoc = PTGToSubnetAssociation(
                        policy_target_group_id=ptg_db.id,
                        subnet_id=subnet
                    )
                    ptg_db.subnets.append(assoc)
            self._process_policy_rule_sets_for_ptg(context, ptg_db, ptg)
            return self._make_policy_target_group_dict(ptg_db)

    @log.log_method_call
    def update_policy_target_group(self, context, policy_target_group_id,
                                   policy_target_group):
        ptg = policy_target_group['policy_target_group']
        with db_api.CONTEXT_WRITER.using(context):
            ptg_db = self._get_policy_target_group(
                context, policy_target_group_id)
            self._process_policy_rule_sets_for_ptg(context, ptg_db, ptg)
            if 'subnets' in ptg:
                # Add/remove associations for changes in subnets.
                new_subnets = set(ptg['subnets'])
                old_subnets = set(subnet.subnet_id
                                  for subnet in ptg_db.subnets)
                for subnet in new_subnets - old_subnets:
                    assoc = PTGToSubnetAssociation(
                        policy_target_group_id=policy_target_group_id,
                        subnet_id=subnet)
                    ptg_db.subnets.append(assoc)
                for subnet in old_subnets - new_subnets:
                    assoc = (
                        context.session.query(
                            PTGToSubnetAssociation).filter_by(
                                policy_target_group_id=policy_target_group_id,
                                subnet_id=subnet).one())
                    ptg_db.subnets.remove(assoc)
                    context.session.delete(assoc)
                # Don't update ptg_db.subnets with subnet IDs.
                del ptg['subnets']
            ptg_db.update(ptg)
            return self._make_policy_target_group_dict(ptg_db)

    @log.log_method_call
    def get_policy_target_groups_count(self, context, filters=None):
        return db_api.get_collection_count(context, PolicyTargetGroupMapping,
                                           filters=filters)

    def _get_policy_target_groups(self, context, filters=None, fields=None,
                           sorts=None, limit=None, marker=None,
                           page_reverse=False):
        plugin = directory.get_plugin()
        marker_obj = db_api.get_marker_obj(plugin, context,
                                           'policy_target_group',
                                           limit, marker)
        return db_api.get_collection(context, PolicyTargetGroupMapping,
                                     self._make_policy_target_group_dict,
                                     filters=filters, fields=fields,
                                     sorts=sorts, limit=limit,
                                     marker_obj=marker_obj,
                                     page_reverse=page_reverse)

    @log.log_method_call
    def get_policy_target_groups(self, context, filters=None, fields=None,
                           sorts=None, limit=None, marker=None,
                           page_reverse=False):
        return self._get_policy_target_groups(
            context, filters, fields, sorts, limit, marker, page_reverse)

    @log.log_method_call
    def create_l2_policy(self, context, l2_policy):
        l2p = l2_policy['l2_policy']
        tenant_id = self._get_tenant_id_for_create(context, l2p)
        with db_api.CONTEXT_WRITER.using(context):
            l2p_db = L2PolicyMapping(id=uuidutils.generate_uuid(),
                                     tenant_id=tenant_id,
                                     name=l2p['name'],
                                     description=l2p['description'],
                                     l3_policy_id=l2p['l3_policy_id'],
                                     network_id=l2p['network_id'],
                                     inject_default_route=l2p.get(
                                         'inject_default_route', True),
                                     shared=l2p.get('shared', False))
            context.session.add(l2p_db)
            return self._make_l2_policy_dict(l2p_db)

    @log.log_method_call
    def get_l2_policies(self, context, filters=None, fields=None,
                        sorts=None, limit=None, marker=None,
                        page_reverse=False):
        plugin = directory.get_plugin()
        marker_obj = db_api.get_marker_obj(plugin, context,
                                           'l2_policy', limit,
                                           marker)
        return db_api.get_collection(context, L2PolicyMapping,
                                     self._make_l2_policy_dict,
                                     filters=filters, fields=fields,
                                     sorts=sorts, limit=limit,
                                     marker_obj=marker_obj,
                                     page_reverse=page_reverse)

    @log.log_method_call
    def get_l2_policies_count(self, context, filters=None):
        return db_api.get_collection_count(context, L2PolicyMapping,
                                           filters=filters)

    @log.log_method_call
    def create_l3_policy(self, context, l3_policy):
        l3p = l3_policy['l3_policy']
        self.validate_ip_pool(l3p.get('ip_pool', None), l3p['ip_version'])
        # TODO(Sumit): Check consistency of explicit subnetpool's
        # address_scopes
        tenant_id = self._get_tenant_id_for_create(context, l3p)
        self.validate_subnet_prefix_length(l3p['ip_version'],
                                           l3p['subnet_prefix_length'],
                                           l3p.get('ip_pool', None))
        with db_api.CONTEXT_WRITER.using(context):
            l3p_db = L3PolicyMapping(
                id=uuidutils.generate_uuid(),
                tenant_id=tenant_id,
                name=l3p['name'],
                ip_version=l3p['ip_version'],
                ip_pool=l3p['ip_pool'],
                subnet_prefix_length=l3p['subnet_prefix_length'],
                description=l3p['description'],
                shared=l3p.get('shared', False))

            self._set_address_scope_for_l3_policy(
                context, l3p_db, l3p.get('address_scope_v4_id'),
                ip_version=4)
            self._set_address_scope_for_l3_policy(
                context, l3p_db, l3p.get('address_scope_v6_id'),
                ip_version=6)
            self._add_subnetpools_to_l3_policy(
                context, l3p_db, l3p.get('subnetpools_v4'), ip_version=4)
            self._add_subnetpools_to_l3_policy(
                context, l3p_db, l3p.get('subnetpools_v6'), ip_version=6)

            if 'routers' in l3p:
                for router in l3p['routers']:
                    assoc = L3PolicyRouterAssociation(
                        l3_policy_id=l3p_db.id,
                        router_id=router
                    )
                    l3p_db.routers.append(assoc)
            if 'external_segments' in l3p:
                self._set_ess_for_l3p(context, l3p_db,
                                      l3p['external_segments'])
            context.session.add(l3p_db)
            return self._make_l3_policy_dict(l3p_db, ip_pool=l3p['ip_pool'])

    @log.log_method_call
    def update_l3_policy(self, context, l3_policy_id, l3_policy):
        l3p = l3_policy['l3_policy']

        if 'address_scope_v4_id' in l3p or 'address_scope_v6_id' in l3p:
            raise AddressScopeUpdateForL3PNotSupported()

        with db_api.CONTEXT_WRITER.using(context):
            l3p_db = self._get_l3_policy(context, l3_policy_id)

            self._update_subnetpools_for_l3_policy(context, l3_policy_id,
                                                   l3p.get('subnetpools_v4'),
                                                   ip_version=4)
            l3p.pop('subnetpools_v4', None)
            self._update_subnetpools_for_l3_policy(context, l3_policy_id,
                                                   l3p.get('subnetpools_v6'),
                                                   ip_version=6)
            l3p.pop('subnetpools_v6', None)

            if 'subnet_prefix_length' in l3p:
                self.validate_subnet_prefix_length(l3p_db.ip_version,
                                                   l3p['subnet_prefix_length'],
                                                   l3p_db.ip_pool)

            if 'routers' in l3p:
                # Add/remove associations for changes in routers.
                new_routers = set(l3p['routers'])
                old_routers = set(router.router_id
                                  for router in l3p_db.routers)
                for router in new_routers - old_routers:
                    assoc = L3PolicyRouterAssociation(
                        l3_policy_id=l3_policy_id, router_id=router)
                    l3p_db.routers.append(assoc)
                for router in old_routers - new_routers:
                    assoc = (context.session.query(L3PolicyRouterAssociation).
                             filter_by(l3_policy_id=l3_policy_id,
                                       router_id=router).
                             one())
                    l3p_db.routers.remove(assoc)
                    context.session.delete(assoc)
                # Don't update l3p_db.routers with router IDs.
                del l3p['routers']
            if 'external_segments' in l3p:
                self._set_ess_for_l3p(context, l3p_db,
                                      l3p['external_segments'])
                del l3p['external_segments']
            l3p_db.update(l3p)
            return self._make_l3_policy_dict(l3p_db)

    @log.log_method_call
    def create_external_segment(self, context, external_segment):
        es = external_segment['external_segment']
        tenant_id = self._get_tenant_id_for_create(context, es)
        with db_api.CONTEXT_WRITER.using(context):
            es_db = ExternalSegmentMapping(
                id=uuidutils.generate_uuid(), tenant_id=tenant_id,
                name=es['name'], description=es['description'],
                shared=es.get('shared', False), ip_version=es['ip_version'],
                cidr=es['cidr'],
                port_address_translation=es['port_address_translation'],
                subnet_id=es['subnet_id'])
            context.session.add(es_db)
            if 'external_routes' in es:
                self._process_segment_ers(context, es_db, es)
            return self._make_external_segment_dict(es_db)

    @log.log_method_call
    def get_external_segments(self, context, filters=None, fields=None,
                              sorts=None, limit=None, marker=None,
                              page_reverse=False):
        plugin = directory.get_plugin()
        marker_obj = db_api.get_marker_obj(plugin, context,
                                           'external_segment', limit,
                                           marker)
        return db_api.get_collection(context, ExternalSegmentMapping,
                                     self._make_external_segment_dict,
                                     filters=filters, fields=fields,
                                     sorts=sorts, limit=limit,
                                     marker_obj=marker_obj,
                                     page_reverse=page_reverse)

    def get_external_segments_count(self, context, filters=None):
        return db_api.get_collection_count(context, ExternalSegmentMapping,
                                           filters=filters)

    @log.log_method_call
    def create_nat_pool(self, context, nat_pool):
        np = nat_pool['nat_pool']
        tenant_id = self._get_tenant_id_for_create(context, np)
        with db_api.CONTEXT_WRITER.using(context):
            np_db = NATPoolMapping(
                id=uuidutils.generate_uuid(), tenant_id=tenant_id,
                name=np['name'], description=np['description'],
                shared=np.get('shared', False), ip_version=np['ip_version'],
                ip_pool=np['ip_pool'],
                external_segment_id=np['external_segment_id'],
                subnet_id=np.get('subnet_id'))
            context.session.add(np_db)
            return self._make_nat_pool_dict(np_db)

    @log.log_method_call
    def get_nat_pools(self, context, filters=None, fields=None,
                      sorts=None, limit=None, marker=None,
                      page_reverse=False):
        plugin = directory.get_plugin()
        marker_obj = db_api.get_marker_obj(plugin, context,
                                           'nat_pool', limit,
                                           marker)
        return db_api.get_collection(context, NATPoolMapping,
                                     self._make_nat_pool_dict,
                                     filters=filters, fields=fields,
                                     sorts=sorts, limit=limit,
                                     marker_obj=marker_obj,
                                     page_reverse=page_reverse)



# Copyright (c) 2018 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.tests import base

from gbpservice.neutron.extensions import cisco_apic as apic_ext


class TestAttributeValidators(base.BaseTestCase):

    def test_validate_apic_vlan(self):
        self.assertIsNone(apic_ext._validate_apic_vlan(None))
        self.assertIsNone(apic_ext._validate_apic_vlan('10'))
        self.assertIsNone(apic_ext._validate_apic_vlan(10))


class TestAttributeConverters(base.BaseTestCase):

    def test_convert_apic_vlan(self):
        self.assertIsInstance(apic_ext.convert_apic_vlan('2'), int)
        self.assertIsInstance(apic_ext.convert_apic_vlan(2), int)
        self.assertIsNone(apic_ext.convert_apic_vlan(None))

    def test_convert_apic_none_to_empty_list(self):
        test_none = None
        expt_list = []
        self.assertItemsEqual(
                apic_ext.convert_apic_none_to_empty_list(
                    test_none), expt_list)

        expt_list = ['value1', 'value2', 'value3']
        test_list_str = "['value1', 'value2', 'value3']"
        self.assertItemsEqual(
                apic_ext.convert_apic_none_to_empty_list(
                    test_list_str), expt_list)

        self.assertItemsEqual(
                apic_ext.convert_apic_none_to_empty_list(
                    expt_list), expt_list)

        expt_list = [{'key1': 'value1'}, {'key2': 'value2'}]
        test_dict_list_str = "[{'key1': 'value1'}, {'key2': 'value2'}]"
        self.assertItemsEqual(
                apic_ext.convert_apic_none_to_empty_list(
                    test_dict_list_str), expt_list)

        self.assertItemsEqual(
                apic_ext.convert_apic_none_to_empty_list(
                    expt_list), expt_list)

    def test_convert_nested_domain_allowed_vlans(self):
        test_dict_str = "{'vlans_list': [2, 3, 4], " + (
                        "'vlan_ranges': [{'start': 6, 'end': 9}, ") + (
                        "{'start': 11, 'end': 14}]}")
        expt_list = [2, 3, 4, 6, 7, 8, 9, 11, 12, 13, 14]
        self.assertItemsEqual(
                apic_ext.convert_nested_domain_allowed_vlans(
                    test_dict_str)['vlans_list'], expt_list)
        test_dict = {'vlans_list': [2, 3, 4],
                     'vlan_ranges': [{'start': 6, 'end': 9},
                                     {'start': 11, 'end': 14}]}
        expt_list = [2, 3, 4, 6, 7, 8, 9, 11, 12, 13, 14]
        self.assertItemsEqual(
                apic_ext.convert_nested_domain_allowed_vlans(
                    test_dict)['vlans_list'], expt_list)
        test_dict = {'vlans_list': ['2', '3', '4', '3'],
                     'vlan_ranges': [{'start': '6', 'end': '9'},
                                     {'start': '11', 'end': '14'},
                                     {'start': '6', 'end': '9'}]}
        expt_list = [2, 3, 4, 6, 7, 8, 9, 11, 12, 13, 14]
        self.assertItemsEqual(
                apic_ext.convert_nested_domain_allowed_vlans(
                    test_dict)['vlans_list'], expt_list)
        self.assertIsNone(apic_ext.convert_nested_domain_allowed_vlans(None))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""add status attributes

Revision ID: 12c1bc8d7026
Revises: 31b399f08b1c
Create Date: 2016-03-08 15:28:57.170563

"""

# revision identifiers, used by Alembic.
revision = '12c1bc8d7026'
down_revision = '31b399f08b1c'

from alembic import op
import sqlalchemy as sa


def upgrade():
    table_names = ['gp_policy_targets', 'gp_policy_target_groups',
                   'gp_l2_policies', 'gp_l3_policies', 'gp_policy_rules',
                   'gp_policy_classifiers', 'gp_policy_actions',
                   'gp_policy_rule_sets', 'gp_nat_pools',
                   'gp_network_service_policies',
                   'gp_external_segments', 'gp_external_policies', 'sc_nodes',
                   'sc_instances', 'sc_specs', 'service_profiles']

    for tname in table_names:
        op.add_column(tname, sa.Column('status', sa.String(length=16),
                                       nullable=True))
        op.add_column(tname, sa.Column('status_details',
                                       sa.String(length=4096), nullable=True))


def downgrade():
    pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Alter HA IP address to Port ID association table to add network id entry
Revision ID: 68fcb81878c5
Revises: f8dc1eb9deaf
Create Date: 2015-10-19 02:08:54.252877
"""

# revision identifiers, used by Alembic.
revision = '68fcb81878c5'
down_revision = 'f8dc1eb9deaf'

from alembic import op
from alembic import util
import sqlalchemy as sa
from sqlalchemy.engine import reflection


def upgrade():

    bind = op.get_bind()
    insp = sa.engine.reflection.Inspector.from_engine(bind)
    if 'apic_ml2_ha_ipaddress_to_port_owner' in insp.get_table_names():

        op.add_column('apic_ml2_ha_ipaddress_to_port_owner',
                      sa.Column('network_id',
                                sa.String(length=36),
                                nullable=False))
        try:
            from gbpservice.neutron.plugins.ml2plus.drivers.apic_aim import (
                data_migrations)

            session = sa.orm.Session(bind=bind, autocommit=True)
            with session.begin(subtransactions=True):
                data_migrations.do_ha_ip_duplicate_entries_removal(session)
                data_migrations.do_ha_ip_network_id_insertion(session)
        except ImportError:
            util.warning("AIM schema present, but failed to import"
                        "AIM libraries - HA IP duplicate entries"
                        "removal not completed.")
        except Exception as e:
            util.warning("Caught exception while migrating data in %s: %s"
                        % ('apic_ml2_ha_ipaddress_to_port_owner', e))

        inspector = reflection.Inspector.from_engine(op.get_bind())
        pk_constraint = inspector.get_pk_constraint(
                'apic_ml2_ha_ipaddress_to_port_owner')
        op.drop_constraint(
            pk_constraint['name'],
            table_name='apic_ml2_ha_ipaddress_to_port_owner',
            type_='primary')
        op.create_primary_key(
            constraint_name='apic_ml2_ha_ipaddress_to_port_owner_pk',
            table_name='apic_ml2_ha_ipaddress_to_port_owner',
            columns=['ha_ip_address', 'network_id'])


def downgrade():
    pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from collections import namedtuple

from oslo_log import log
import sqlalchemy as sa
from sqlalchemy.ext import baked

from gbpservice.neutron.db.grouppolicy.extensions import (
    apic_auto_ptg_db as auto_ptg_db)
from gbpservice.neutron.db.grouppolicy.extensions import (
    apic_segmentation_label_db as seg_label_db)
from gbpservice.neutron.db.grouppolicy import group_policy_mapping_db as gpmdb
from gbpservice.neutron.plugins.ml2plus.drivers.apic_aim import (
    constants as md_const)


LOG = log.getLogger(__name__)

BAKERY = baked.bakery(_size_alert=lambda c: LOG.warning(
    "sqlalchemy baked query cache size exceeded in %s", __name__))

EndpointPtInfo = namedtuple(
    'EndpointPtInfo',
    ['pt_id',
     'ptg_id',
     'apg_id',
     'inject_default_route',
     'l3p_project_id',
     'is_auto_ptg'])


class AIMMappingRPCMixin(object):
    # The query_endpoint_rpc_info and update_endpoint_rpc_details
    # methods below are called by the apic_aim mechanism driver while
    # handling the request_endpoint_details (aka get_gbp_details) RPC
    # from the agent.

    def query_endpoint_rpc_info(self, session, info):
        # This method is called within a transaction from the apic_aim
        # MD's request_endpoint_details RPC handler to retrieve GBP
        # state needed to build the RPC response, after the info param
        # has already been populated with the data available within
        # Neutron itself.

        # Query for all needed scalar (non-list) state for the
        # policies associated with the port, and make sure the port is
        # owned by a policy target before continuing.
        pt_infos = self._query_pt_info(
            session, info['port_info'].port_id)
        if not pt_infos:
            return

        # A list was returned by the PT info query, like all the other
        # endpoint RPC queries, here and in the mechanism
        # driver. Currently, there will be at most a single item in
        # this list, but a join may later be added to this query in
        # order to eliminate another query's round-trip to the DB
        # server, resulting in multiple rows being returned. For now,
        # we just need that single row.
        pt_info = pt_infos[0]
        info['gbp_pt_info'] = pt_info

        # Query for policy target's segmentation labels.
        info['gbp_segmentation_labels'] = self._query_segmentation_labels(
            session, pt_info.pt_id)

    def _query_pt_info(self, session, port_id):
        query = BAKERY(lambda s: s.query(
            gpmdb.PolicyTargetMapping.id,
            gpmdb.PolicyTargetMapping.policy_target_group_id,
            gpmdb.PolicyTargetGroupMapping.application_policy_group_id,
            gpmdb.L2PolicyMapping.inject_default_route,
            gpmdb.L3PolicyMapping.project_id,
            auto_ptg_db.ApicAutoPtgDB.is_auto_ptg,
        ))
        query += lambda q: q.join(
            gpmdb.PolicyTargetGroupMapping,
            gpmdb.PolicyTargetGroupMapping.id ==
            gpmdb.PolicyTargetMapping.policy_target_group_id)
        query += lambda q: q.join(
            gpmdb.L2PolicyMapping,
            gpmdb.L2PolicyMapping.id ==
            gpmdb.PolicyTargetGroupMapping.l2_policy_id)
        query += lambda q: q.join(
            gpmdb.L3PolicyMapping,
            gpmdb.L3PolicyMapping.id ==
            gpmdb.L2PolicyMapping.l3_policy_id)
        query += lambda q: q.outerjoin(
            auto_ptg_db.ApicAutoPtgDB,
            auto_ptg_db.ApicAutoPtgDB.policy_target_group_id ==
            gpmdb.PolicyTargetMapping.policy_target_group_id)
        query += lambda q: q.filter(
            gpmdb.PolicyTargetMapping.port_id == sa.bindparam('port_id'))
        return [EndpointPtInfo._make(row) for row in
                query(session).params(
                    port_id=port_id)]

    def _query_segmentation_labels(self, session, pt_id):
        query = BAKERY(lambda s: s.query(
            seg_label_db.ApicSegmentationLabelDB.segmentation_label))
        query += lambda q: q.filter(
            seg_label_db.ApicSegmentationLabelDB.policy_target_id ==
            sa.bindparam('pt_id'))
        return [x for x, in query(session).params(
            pt_id=pt_id)]

    def update_endpoint_rpc_details(self, info, details):
        # This method is called outside a transaction from the
        # apic_aim MD's request_endpoint_details RPC handler to add or
        # update details within the RPC response, using data stored in
        # info by query_endpoint_rpc_info.

        # First, make sure the port is owned by a PolicyTarget before
        # continuing.
        pt_info = info.get('gbp_pt_info')
        if not pt_info:
            return
        gbp_details = details['gbp_details']

        # Replace EPG identity if not auto_ptg.
        if not pt_info.is_auto_ptg:
            gbp_details['app_profile_name'] = (
                self.name_mapper.application_policy_group(
                    None, pt_info.apg_id) if pt_info.apg_id
                else self.aim_mech_driver.ap_name)
            gbp_details['endpoint_group_name'] = pt_info.ptg_id
            gbp_details['ptg_tenant'] = (
                self.name_mapper.project(None, pt_info.l3p_project_id))

        # Update subnet gateway_ip and default_routes if needed.
        if not pt_info.inject_default_route:
            for subnet in gbp_details['subnets']:
                del subnet['gateway_ip']
                subnet['host_routes'] = [
                    r for r in subnet['host_routes']
                    if r['destination'] not in
                    [md_const.IPV4_ANY_CIDR, md_const.IPV4_METADATA_CIDR]]

        # Add segmentation labels.
        gbp_details['segmentation_labels'] = (
            info.get('gbp_segmentation_labels'))

        # REVISIT: If/when support for the proxy_group extension is
        # added to the aim_mapping PD, update promiscuous_mode to True
        # if this PT has a cluster_id that identifies a different PT
        # whose group_default_gateway set.



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Table for AIM address-scope extension attribute

Revision ID: 8cd6d095d7d3
Revises: 3e1f67cf951b
Create Date: 2017-01-05 19:17:45.088969

"""

# revision identifiers, used by Alembic.
revision = '8cd6d095d7d3'
down_revision = '3e1f67cf951b'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'apic_aim_addr_scope_extensions',
        sa.Column('address_scope_id', sa.String(36), nullable=False),
        sa.Column('vrf_dn', sa.String(1024)),
        sa.ForeignKeyConstraint(['address_scope_id'], ['address_scopes.id'],
                                name='apic_aim_addr_scope_extn_fk_addr_scope',
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('address_scope_id')
    )


def downgrade():
    pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron.tests import base

from gbpservice.contrib.nfp.configurator.lib import demuxer as demuxer_lib
from gbpservice.contrib.nfp.configurator.modules import configurator as cfgr
from gbpservice.contrib.tests.unit.nfp.configurator.test_data import (
                                                        fw_test_data as fo)


class ConfiguratorRpcManagerTestCase(base.BaseTestCase):
    """ Tests RPC manager class of configurator

    """

    def __init__(self, *args, **kwargs):
        super(ConfiguratorRpcManagerTestCase, self).__init__(*args, **kwargs)
        self.fo = fo.FakeObjects()

    @mock.patch(__name__ + '.fo.FakeObjects.conf')
    @mock.patch(__name__ + '.fo.FakeObjects.sc')
    def _get_ConfiguratorRpcManager_object(self, sc, conf):
        """ Retrieves RPC manager object of configurator.

        :param sc: mocked service controller object of process model framework
        :param conf: mocked OSLO configuration file

        Returns: object of configurator's RPC manager.

        """

        cm = cfgr.ConfiguratorModule(sc)
        demuxer = demuxer_lib.ServiceAgentDemuxer()
        rpc_mgr = cfgr.ConfiguratorRpcManager(sc, cm, conf, demuxer)
        return sc, rpc_mgr

    def _test_network_function_device_config(self, operation,
                                             method, batch=False):
        """ Tests generic config APIs

        :param operation: create/delete
        :param method: CONFIGURE_ROUTES/CLEAR_ROUTES/
        CONFIGURE_INTERFACES/CLEAR_INTERFACES
        :param batch: True or False. Indicates if the
        request is a batch request

        Returns: none

        """

        _, rpc_mgr = self._get_ConfiguratorRpcManager_object()
        agent = mock.Mock()

        request_data = {'batch': {
                'request_data_actual': (
                            self.fo.fake_request_data_generic_bulk()),
                'request_data_expected': (
                            self.fo.fake_request_data_generic_bulk())},
                        'single': {
                'request_data_actual': (
                            (self.fo.fake_request_data_generic_single(
                                                                routes=True)
                             if 'ROUTES' in method
                             else self.fo.fake_request_data_generic_single())),
                'request_data_expected': (
                            (self.fo.fake_request_data_generic_single(
                                                                routes=True)
                             if 'ROUTES' in method
                             else self.fo.fake_request_data_generic_single()))}
                        }
        if batch:
            request_data_actual, _ = (list(request_data['batch'].values()))
        else:
            request_data_actual, _ = (list(request_data['single'].values()))

        with mock.patch.object(rpc_mgr,
                               '_get_service_agent_instance',
                               return_value=agent), (
             mock.patch.object(agent, 'process_request')) as mock_request:

            if operation == 'create':
                rpc_mgr.create_network_function_device_config(
                                    self.fo.context, request_data_actual)
            elif operation == 'delete':
                rpc_mgr.delete_network_function_device_config(
                                    self.fo.context, request_data_actual)

            notification_data = dict()
            sa_req_list = self.fo.fake_sa_req_list()

            response_data = {'single': {'routes': [sa_req_list[1]],
                                        'interfaces': [sa_req_list[0]]},
                             'batch': sa_req_list}

            if batch:
                data = response_data['batch']
                if operation == 'delete':
                    data[0]['method'] = 'clear_interfaces'
                    data[1]['method'] = 'clear_routes'
            else:
                data = response_data['single'][method.split('_')[1].lower()]
                if operation == 'delete':
                    data[0]['method'] = data[0]['method'].replace(
                                                    'configure', 'clear', 1)
            for agent_info in data:
                agent_info['agent_info']['context']['nf_id'] = 'nf_id'
                agent_info['agent_info']['context']['nfi_id'] = 'nfi_id'
            mock_request.assert_called_with(data,
                                            notification_data)

    def _test_network_function_config(self, operation):
        """ Tests firewall APIs

        :param operation: CREATE_FIREWALL/UPDATE_FIREWALL/DELETE_FIREWALL

        Returns: none

        """

        _, rpc_mgr = self._get_ConfiguratorRpcManager_object()
        agent = mock.Mock()
        method = {'CREATE': 'create_network_function_config',
                  'UPDATE': 'update_network_function_config',
                  'DELETE': 'delete_network_function_config'}
        request_data = self.fo.fake_request_data_fw()
        with mock.patch.object(rpc_mgr,
                               '_get_service_agent_instance',
                               return_value=agent), (
             mock.patch.object(agent, 'process_request')) as mock_request:

            getattr(rpc_mgr, method[operation.split('_')[0]])(
                                self.fo.firewall_api_context(), request_data)

            notification_data = dict()
            data = self.fo.fake_sa_req_list_fw()
            if 'UPDATE' in operation:
                data[0]['method'] = data[0]['method'].replace(
                                                    'create', 'update', 1)
            elif 'DELETE' in operation:
                data[0]['method'] = data[0]['method'].replace(
                                                    'create', 'delete', 1)

            mock_request.assert_called_with(data,
                                            notification_data)

    def _test_notifications(self):
        """ Tests response path notification  APIs

        Returns: none

        """

        sc, rpc_mgr = self._get_ConfiguratorRpcManager_object()

        events = fo.FakeEventGetNotifications()
        with mock.patch.object(sc, 'get_stashed_events',
                               return_value=[events]):

            return_value = rpc_mgr.get_notifications('context')

            expected_value = [events.data]
            self.assertEqual(return_value, expected_value)

    def test_configure_routes_generic_api(self):
        """ Implements test case for configure routes API

        Returns: none

        """

        method = "CONFIGURE_ROUTES"
        operation = 'create'
        self._test_network_function_device_config(operation, method)

    def test_clear_routes_generic_api(self):
        """ Implements test case for clear routes API

        Returns: none

        """

        method = "CLEAR_ROUTES"
        operation = 'delete'
        self._test_network_function_device_config(operation, method)

    def test_configure_interfaces_generic_api(self):
        """ Implements test case for configure interfaces API

        Returns: none

        """

        method = "CONFIGURE_INTERFACES"
        operation = 'create'
        self._test_network_function_device_config(operation, method)

    def test_clear_interfaces_generic_api(self):
        """ Implements test case for clear interfaces API

        Returns: none

        """

        method = "CLEAR_INTERFACES"
        operation = 'delete'
        self._test_network_function_device_config(operation, method)

    def test_configure_bulk_generic_api(self):
        """ Implements test case for bulk configure request API

        Returns: none

        """

        method = "PROCESS_BATCH"
        operation = 'create'
        self._test_network_function_device_config(operation, method, True)

    def test_clear_bulk_generic_api(self):
        """ Implements test case for bulk clear request API

        Returns: none

        """

        method = "PROCESS_BATCH"
        operation = 'delete'
        self._test_network_function_device_config(operation, method, True)

    def test_network_function_create_api(self):
        """ Implements test case for create firewall API

        Returns: none

        """

        self._test_network_function_config('CREATE_FIREWALL')

    def test_network_function_update_api(self):
        """ Implements test case for update firewall API

        Returns: none

        """

        self._test_network_function_config('UPDATE_FIREWALL')

    def test_network_function_delete_api(self):
        """ Implements test case for delete firewall API

        Returns: none

        """

        self._test_network_function_config('DELETE_FIREWALL')

    def test_get_notifications_generic_configurator_api(self):
        """ Implements test case for get notifications API
        of configurator

        Returns: none

        """

        self._test_notifications()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from gbpservice.neutron.db import api as db_api
from gbpservice.neutron.db.grouppolicy.extensions import (
    apic_allowed_vm_name_db as db)
from gbpservice.neutron.tests.unit.services.grouppolicy import (
    test_extension_driver_api as test_ext_base)


class ExtensionDriverTestCase(test_ext_base.ExtensionDriverTestBase):
    _extension_drivers = ['apic_allowed_vm_name']
    _extension_path = None

    def test_l3p_lifecycle(self):
        l3p = self.create_l3_policy(name='myl3')['l3_policy']
        self.assertEqual([], l3p['allowed_vm_names'])
        l3p = self.show_l3_policy(
            l3p['id'], expected_res_status=200)['l3_policy']
        self.assertEqual([], l3p['allowed_vm_names'])
        self.delete_l3_policy(l3p['id'], tenant_id=l3p['tenant_id'],
                              expected_res_status=204)

        allowed_vm_names = []
        l3p = self.create_l3_policy(
            name='myl3',
            allowed_vm_names=allowed_vm_names)['l3_policy']
        self.assertItemsEqual(allowed_vm_names, l3p['allowed_vm_names'])
        l3p = self.show_l3_policy(
            l3p['id'], expected_res_status=200)['l3_policy']
        self.assertItemsEqual([], l3p['allowed_vm_names'])
        self.delete_l3_policy(l3p['id'], tenant_id=l3p['tenant_id'],
                              expected_res_status=204)

        allowed_vm_names = ['safe_vm*', '^secure_vm*']
        l3p = self.create_l3_policy(
            name='myl3',
            allowed_vm_names=allowed_vm_names)['l3_policy']
        self.assertItemsEqual(allowed_vm_names, l3p['allowed_vm_names'])
        l3p = self.show_l3_policy(
            l3p['id'], expected_res_status=200)['l3_policy']
        self.assertItemsEqual(allowed_vm_names, l3p['allowed_vm_names'])

        allowed_vm_names = ['good_vm*', '^ok_vm*', 'safe_vm*']
        l3p = self.update_l3_policy(
            l3p['id'], allowed_vm_names=allowed_vm_names,
            expected_res_status=200)['l3_policy']
        self.assertItemsEqual(allowed_vm_names, l3p['allowed_vm_names'])
        l3p = self.show_l3_policy(
            l3p['id'], expected_res_status=200)['l3_policy']
        self.assertItemsEqual(allowed_vm_names, l3p['allowed_vm_names'])

        allowed_vm_names = []
        l3p = self.update_l3_policy(
            l3p['id'], allowed_vm_names=allowed_vm_names,
            expected_res_status=200)['l3_policy']
        self.assertItemsEqual(allowed_vm_names, l3p['allowed_vm_names'])
        l3p = self.show_l3_policy(
            l3p['id'], expected_res_status=200)['l3_policy']
        self.assertItemsEqual(allowed_vm_names, l3p['allowed_vm_names'])

        allowed_vm_names = ['^ok_vm*']
        l3p = self.update_l3_policy(
            l3p['id'], allowed_vm_names=allowed_vm_names,
            expected_res_status=200)['l3_policy']
        self.assertItemsEqual(allowed_vm_names, l3p['allowed_vm_names'])
        l3p = self.show_l3_policy(
            l3p['id'], expected_res_status=200)['l3_policy']
        self.assertItemsEqual(allowed_vm_names, l3p['allowed_vm_names'])

        self.delete_l3_policy(l3p['id'], tenant_id=l3p['tenant_id'],
                              expected_res_status=204)
        session = db_api.get_reader_session()
        rows = (session.query(db.ApicAllowedVMNameDB).filter_by(
                l3_policy_id=l3p['id']).all())
        self.assertEqual([], rows)



# Copyright (c) 2017 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# The following are imported at the beginning to ensure that the
# patches are applied before any of the modules save a reference to
# the functions being patched. The order is also important.
from gbpservice.neutron.extensions import patch  # noqa

from gbpservice.neutron.plugins.ml2plus import patch_neutron  # noqa

from oslo_db.sqlalchemy import utils as sa_utils

# REVISIT: Remove this as soon as possible.
if not hasattr(sa_utils, '_get_unique_keys'):
    sa_utils._get_unique_keys = sa_utils.get_unique_keys


from gbpservice.neutron.plugins.ml2plus.drivers.apic_aim import cache


# The following is to avoid excessive logging in the UTs
cache.LOG.warning = cache.LOG.info



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.from gbpservice.neutron.nsf.core import main

from gbpservice.nfp.core import context as nfp_context
from gbpservice.nfp.core import event
from gbpservice.nfp.core import module as nfp_api

"""An example NFP module used for UT.

    Implements a sample NFP module used by UT code.
    Event handlers for the events generated by UT
    code are implemented here.
"""


class EventsHandler(nfp_api.NfpEventHandler):

    def __init__(self, controller):
        self.controller = controller

    def handle_event(self, event):
        event.context['log_context']['namespace'] = event.desc.target
        nfp_context.init(event.context)

        if event.id == 'TEST_EVENT_ACK_FROM_WORKER':
            self.controller.event_ack_handler_cb_obj.set()

        if event.id == 'TEST_POST_EVENT_FROM_WORKER':
            self.controller.post_event_worker_wait_obj.set()

        if event.id == 'TEST_POLL_EVENT_FROM_WORKER':
            self.controller.poll_event_worker_wait_obj.set()
            self.controller.poll_event(event, spacing=1)

        if event.id == 'TEST_POLL_EVENT_CANCEL_FROM_WORKER':
            self.controller.poll_event_worker_wait_obj.set()
            self.controller.poll_event(event, spacing=1, max_times=2)

    def handle_poll_event(self, event):
        if event.id == 'TEST_POLL_EVENT_FROM_WORKER':
            self.controller.poll_event_poll_wait_obj.set()
        if event.id == 'TEST_POLL_EVENT_CANCEL_FROM_WORKER':
            self.controller.poll_event_poll_wait_obj.set()

    def event_cancelled(self, event, reason):
        if event.id == 'TEST_POLL_EVENT_CANCEL_FROM_WORKER':
            if reason == 'MAX_TIMED_OUT':
                self.controller.poll_event_poll_cancel_wait_obj.set()

    @nfp_api.poll_event_desc(event='POLL_EVENT_DECORATOR', spacing=2)
    def handle_poll_event_desc(self, event):
        pass


def nfp_module_post_init(controller, conf):
    if hasattr(controller, 'nfp_module_post_init_wait_obj'):
        controller.nfp_module_post_init_wait_obj.set()


def nfp_module_init(controller, conf):
    if hasattr(controller, 'nfp_module_init_wait_obj'):
        controller.nfp_module_init_wait_obj.set()

    evs = [
        event.Event(id='EVENT_1', handler=EventsHandler(controller)),
        event.Event(id='EVENT_LOAD_1', handler=EventsHandler(controller)),
        event.Event(id='EVENT_LOAD_2', handler=EventsHandler(controller)),
        event.Event(id='EVENT_LOAD_3', handler=EventsHandler(controller)),
        event.Event(id='EVENT_LOAD_4', handler=EventsHandler(controller)),
        event.Event(id='EVENT_LOAD_5', handler=EventsHandler(controller)),
        event.Event(id='EVENT_LOAD_6', handler=EventsHandler(controller)),
        event.Event(id='EVENT_LOAD_7', handler=EventsHandler(controller)),
        event.Event(id='EVENT_LOAD_8', handler=EventsHandler(controller)),
        event.Event(id='EVENT_LOAD_9', handler=EventsHandler(controller)),
        event.Event(id='EVENT_LOAD_10', handler=EventsHandler(controller)),
        event.Event(id='SEQUENCE_EVENT_1', handler=EventsHandler(controller)),
        event.Event(id='SEQUENCE_EVENT_2', handler=EventsHandler(controller)),
        event.Event(id='POLL_EVENT', handler=EventsHandler(controller)),
        event.Event(id='POLL_EVENT_DECORATOR',
                    handler=EventsHandler(controller)),
        event.Event(id='POLL_EVENT_WITHOUT_SPACING',
                    handler=EventsHandler(controller)),
        event.Event(id='TEST_EVENT_ACK_FROM_WORKER',
                    handler=EventsHandler(controller)),
        event.Event(id='TEST_POST_EVENT_FROM_WORKER',
                    handler=EventsHandler(controller)),
        event.Event(id='TEST_POLL_EVENT_FROM_WORKER',
                    handler=EventsHandler(controller)),
        event.Event(id='TEST_POLL_EVENT_CANCEL_FROM_WORKER',
                    handler=EventsHandler(controller))
    ]
    controller.register_events(evs)



# Copyright 2014 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""ep_redirect_action
"""

# revision identifiers, used by Alembic.
revision = '2f3834ea746b'
down_revision = '97ac978118c6'

from alembic import op
from sqlalchemy.engine import reflection


def upgrade(active_plugins=None, options=None):
    inspector = reflection.Inspector.from_engine(op.get_bind())
    fk_name = [fk['name'] for fk in
               inspector.get_foreign_keys('gpm_ptgs_servicechain_mapping')
               if 'consumer_ptg_id' in fk['constrained_columns']]
    op.drop_constraint(fk_name[0], 'gpm_ptgs_servicechain_mapping',
                       'foreignkey')


def downgrade(active_plugins=None, options=None):
    # Downgrade would require deleting external policy entries resulting in
    # data loss. Also all the existing SCIs should be inspected and deleted
    # whenever they refer to external policies.
    pass



# Copyright 2014 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Group Policy Mapping DB (gpm_db_1)

Revision ID: 53de98f7a066
Revises: ab64381ee820
Create Date: 2014-07-24 13:58:14.716984

"""

# revision identifiers, used by Alembic.
revision = '53de98f7a066'
down_revision = 'ab64381ee820'


from alembic import op
import sqlalchemy as sa


def upgrade():

    op.create_table(
        'gp_ptg_to_subnet_associations',
        sa.Column('policy_target_group_id', sa.String(length=36),
                  nullable=False),
        sa.Column('subnet_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['policy_target_group_id'],
                                ['gp_policy_target_groups.id']),
        sa.ForeignKeyConstraint(['subnet_id'], ['subnets.id']),
        sa.PrimaryKeyConstraint('policy_target_group_id', 'subnet_id')
    )

    op.create_table(
        'gp_l3_policy_router_associations',
        sa.Column('l3_policy_id', sa.String(length=36), nullable=False),
        sa.Column('router_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['l3_policy_id'], ['gp_l3_policies.id']),
        sa.ForeignKeyConstraint(['router_id'], ['routers.id']),
        sa.PrimaryKeyConstraint('l3_policy_id', 'router_id')
    )

    op.add_column(
        'gp_policy_target_groups',
        sa.Column('type', sa.String(length=15), nullable=True)
    )

    op.add_column(
        'gp_policy_targets',
        sa.Column('type', sa.String(length=15), nullable=True)
    )

    op.add_column(
        'gp_l2_policies',
        sa.Column('type', sa.String(length=15), nullable=True)
    )

    op.add_column(
        'gp_l3_policies',
        sa.Column('type', sa.String(length=15), nullable=True)
    )

    op.add_column(
        'gp_policy_targets',
        sa.Column('port_id', sa.String(length=36), nullable=True)
    )
    op.create_unique_constraint(None, 'gp_policy_targets', ['port_id'])
    op.create_foreign_key('gp_policy_targets_ibfk_2',
                          source_table='gp_policy_targets',
                          referent_table='ports',
                          local_cols=['port_id'],
                          remote_cols=['id'], ondelete='SET NULL')

    op.add_column(
        'gp_l2_policies',
        sa.Column('network_id', sa.String(length=36), nullable=True)
    )
    op.create_unique_constraint(None, 'gp_l2_policies', ['network_id'])
    op.create_foreign_key('gp_l2_policies_ibfk_2',
                          source_table='gp_l2_policies',
                          referent_table='networks',
                          local_cols=['network_id'],
                          remote_cols=['id'])


def downgrade():

    op.drop_constraint('gp_l2_policies_ibfk_2', 'gp_l2_policies', 'foreignkey')
    op.drop_column('gp_l2_policies', 'network_id')
    op.drop_constraint('gp_policy_targets_ibfk_2', 'gp_policy_targets',
                       'foreignkey')
    op.drop_column('gp_policy_targets', 'port_id')
    op.drop_column('gp_l3_policies', 'type')
    op.drop_column('gp_l2_policies', 'type')
    op.drop_column('gp_policy_targets', 'type')
    op.drop_column('gp_policy_target_groups', 'type')
    op.drop_table('gp_l3_policy_router_associations')
    op.drop_table('gp_ptg_to_subnet_associations')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""nested domain attributes in cisco_apic extension

Revision ID: 4967af35820f
Revises: 1c564e737f9f
Create Date: 2018-04-19 14:18:11.909757

"""

# revision identifiers, used by Alembic.
revision = '4967af35820f'
down_revision = '1c564e737f9f'

from alembic import op
import sqlalchemy as sa


def upgrade():

    op.add_column('apic_aim_network_extensions',
                  sa.Column('nested_domain_name', sa.String(1024),
                      nullable=True))
    op.add_column('apic_aim_network_extensions',
                  sa.Column('nested_domain_type', sa.String(1024),
                      nullable=True))
    op.add_column('apic_aim_network_extensions',
                  sa.Column('nested_domain_infra_vlan', sa.Integer,
                      nullable=True))
    op.add_column('apic_aim_network_extensions',
                  sa.Column('nested_domain_service_vlan', sa.Integer,
                      nullable=True))
    op.add_column('apic_aim_network_extensions',
                  sa.Column('nested_domain_node_network_vlan', sa.Integer,
                      nullable=True))

    op.create_table(
        'apic_aim_network_nested_domain_allowed_vlans',
        sa.Column('vlan', sa.Integer, nullable=False),
        sa.PrimaryKeyConstraint('vlan'),
        sa.Column('network_id', sa.String(36), nullable=False),
        sa.ForeignKeyConstraint(['network_id'], ['networks.id'],
                                name='apic_aim_network_nested_extn_fk_network',
                                ondelete='CASCADE')
    )


def downgrade():
    pass



# Copyright (c) 2016 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from collections import defaultdict

from neutron.db import models_v2
from neutron_lib.db import model_base
from oslo_log import log
import sqlalchemy as sa
from sqlalchemy.ext import baked
from sqlalchemy import orm
from sqlalchemy.sql.expression import true

from gbpservice.neutron.extensions import cisco_apic
from gbpservice.neutron.extensions import cisco_apic_l3

LOG = log.getLogger(__name__)

BAKERY = baked.bakery(_size_alert=lambda c: LOG.warning(
    "sqlalchemy baked query cache size exceeded in %s", __name__))


class PortExtensionErspanDb(model_base.BASEV2):

    __tablename__ = 'apic_aim_port_erspan_configurations'

    port_id = sa.Column(
        sa.String(36), sa.ForeignKey('ports.id', ondelete="CASCADE"),
        primary_key=True)
    dest_ip = sa.Column(sa.String(64), primary_key=True)
    flow_id = sa.Column(sa.Integer, primary_key=True)
    direction = sa.Column(sa.Enum('in', 'out', 'both'),
                          default='both', primary_key=True)
    port = orm.relationship(models_v2.Port,
                            backref=orm.backref(
                                'aim_extension_erspan_configs',
                                uselist=True,
                                lazy='joined', cascade='delete'))


class NetworkExtensionDb(model_base.BASEV2):

    __tablename__ = 'apic_aim_network_extensions'

    network_id = sa.Column(
        sa.String(36), sa.ForeignKey('networks.id', ondelete="CASCADE"),
        primary_key=True)
    external_network_dn = sa.Column(sa.String(1024))
    bridge_domain_dn = sa.Column(sa.String(1024))
    nat_type = sa.Column(sa.Enum('distributed', 'edge', ''))
    svi = sa.Column(sa.Boolean)
    bgp_enable = sa.Column(sa.Boolean, default=False, nullable=False)
    bgp_type = sa.Column(sa.Enum('default_export', ''),
                         default='default_export',
                         nullable=False)
    bgp_asn = sa.Column(sa.String(64), default='0', nullable=False)
    policy_enforcement_pref = sa.Column(sa.Enum('unenforced', 'enforced', ''),
                         default='unenforced',
                         nullable=False)

    network = orm.relationship(models_v2.Network,
                               backref=orm.backref(
                                   'aim_extension_mapping', lazy='joined',
                                   uselist=False, cascade='delete'))
    nested_domain_name = sa.Column(sa.String(1024), nullable=True)
    nested_domain_type = sa.Column(sa.String(1024), nullable=True)
    nested_domain_infra_vlan = sa.Column(sa.Integer, nullable=True)
    nested_domain_service_vlan = sa.Column(sa.Integer, nullable=True)
    nested_domain_node_network_vlan = sa.Column(sa.Integer, nullable=True)
    multi_ext_nets = sa.Column(sa.Boolean, default=False, nullable=False)


class NetworkExtensionCidrDb(model_base.BASEV2):

    __tablename__ = 'apic_aim_network_external_cidrs'

    network_id = sa.Column(
        sa.String(36), sa.ForeignKey('networks.id', ondelete="CASCADE"),
        primary_key=True)
    cidr = sa.Column(sa.String(64), primary_key=True)
    network = orm.relationship(models_v2.Network,
                               backref=orm.backref(
                                   'aim_extension_cidr_mapping', lazy='joined',
                                   uselist=True, cascade='delete'))


class NetworkExtNestedDomainAllowedVlansDb(model_base.BASEV2):

    __tablename__ = 'apic_aim_network_nested_domain_allowed_vlans'

    # There is a single pool of VLANs for an APIC
    vlan = sa.Column(sa.Integer(), primary_key=True)
    network_id = sa.Column(
        sa.String(36), sa.ForeignKey('networks.id', ondelete="CASCADE"))
    network = orm.relationship(models_v2.Network,
                               backref=orm.backref(
                                   'aim_extension_domain_mapping',
                                   uselist=True,
                                   lazy='joined', cascade='delete'))


class NetworkExtExtraContractDb(model_base.BASEV2):

    __tablename__ = 'apic_aim_network_extra_contracts'

    network_id = sa.Column(
        sa.String(36), sa.ForeignKey('networks.id', ondelete="CASCADE"))
    contract_name = sa.Column(sa.String(64), primary_key=True)
    provides = sa.Column(sa.Boolean, primary_key=True)
    network = orm.relationship(models_v2.Network,
                               backref=orm.backref(
                                   'aim_extension_extra_contract_mapping',
                                   uselist=True,
                                   lazy='joined', cascade='delete'))


class NetworkExtEpgContractMasterDb(model_base.BASEV2):

    __tablename__ = 'apic_aim_network_epg_contract_masters'

    network_id = sa.Column(
        sa.String(36), sa.ForeignKey('networks.id', ondelete="CASCADE"))
    app_profile_name = sa.Column(sa.String(64), primary_key=True)
    name = sa.Column(sa.String(64), primary_key=True)
    network = orm.relationship(models_v2.Network,
                               backref=orm.backref(
                                   'aim_extension_epg_contract_masters',
                                   uselist=True,
                                   lazy='joined', cascade='delete'))


class NetworkExtensionNoNatCidrsDb(model_base.BASEV2):

    __tablename__ = 'apic_aim_network_no_nat_cidrs'

    network_id = sa.Column(
        sa.String(36), sa.ForeignKey('networks.id', ondelete="CASCADE"),
        primary_key=True)
    cidr = sa.Column(sa.String(64), primary_key=True)
    network = orm.relationship(models_v2.Network,
                               backref=orm.backref(
                                   'aim_extension_no_nat_cidrs_mapping',
                                   lazy='joined', uselist=True,
                                   cascade='delete'))


class SubnetExtensionDb(model_base.BASEV2):

    __tablename__ = 'apic_aim_subnet_extensions'

    subnet_id = sa.Column(
        sa.String(36), sa.ForeignKey('subnets.id', ondelete="CASCADE"),
        primary_key=True)
    snat_host_pool = sa.Column(sa.Boolean)
    active_active_aap = sa.Column(sa.Boolean)
    snat_subnet_only = sa.Column(sa.Boolean)
    epg_subnet = sa.Column(sa.Boolean)
    subnet = orm.relationship(models_v2.Subnet,
                              backref=orm.backref(
                                  'aim_extension_mapping', lazy='joined',
                                  uselist=False, cascade='delete'))


class RouterExtensionContractDb(model_base.BASEV2):

    __tablename__ = 'apic_aim_router_external_contracts'

    router_id = sa.Column(
        sa.String(36), sa.ForeignKey('routers.id', ondelete="CASCADE"),
        primary_key=True)
    contract_name = sa.Column(sa.String(64), primary_key=True)
    provides = sa.Column(sa.Boolean, primary_key=True)


class ExtensionDbMixin(object):

    def _set_if_not_none(self, res_dict, res_attr, db_attr):
        if db_attr is not None:
            res_dict[res_attr] = db_attr

    def get_port_extn_db(self, session, port_id):
        return self.get_port_extn_db_bulk(session, [port_id]).get(
            port_id, {})

    def get_port_extn_db_bulk(self, session, port_ids):
        if not port_ids:
            return {}

        query = BAKERY(lambda s: s.query(
            PortExtensionErspanDb))
        query += lambda q: q.filter(
            PortExtensionErspanDb.port_id.in_(
                sa.bindparam('port_ids', expanding=True)))
        db_erspans = query(session).params(
            port_ids=port_ids).all()

        erspans_by_port_id = {}
        for db_erspan in db_erspans:
            erspans_by_port_id.setdefault(db_erspan.port_id, []).append(
                db_erspan)

        result = {}
        for db_obj in db_erspans:
            port_id = db_obj.port_id
            result.setdefault(port_id, self.make_port_extn_db_conf_dict(
                erspans_by_port_id.get(port_id, [])))
        return result

    def make_port_extn_db_conf_dict(self, db_erspans):
        port_res = {}
        db_obj = db_erspans
        if db_obj:
            def _db_to_dict(db_obj):
                ed = {cisco_apic.ERSPAN_DEST_IP: db_obj.dest_ip,
                      cisco_apic.ERSPAN_FLOW_ID: db_obj.flow_id,
                      cisco_apic.ERSPAN_DIRECTION: db_obj.direction}
                return ed
            port_res[cisco_apic.ERSPAN_CONFIG] = [_db_to_dict(e)
                                                  for e in db_erspans]
        return port_res

    def set_port_extn_db(self, session, port_id, res_dict):
        with session.begin(subtransactions=True):
            if cisco_apic.ERSPAN_CONFIG in res_dict:
                self._update_dict_attr(
                        session, PortExtensionErspanDb,
                        (cisco_apic.ERSPAN_DEST_IP,
                         cisco_apic.ERSPAN_FLOW_ID,
                         cisco_apic.ERSPAN_DIRECTION
                         ),
                        res_dict[cisco_apic.ERSPAN_CONFIG],
                        port_id=port_id)

    def get_network_extn_db(self, session, network_id):
        return self.get_network_extn_db_bulk(session, [network_id]).get(
            network_id, {})

    def get_network_extn_db_bulk(self, session, network_ids):
        if not network_ids:
            return {}

        query = BAKERY(lambda s: s.query(
            NetworkExtensionDb))
        query += lambda q: q.filter(
            NetworkExtensionDb.network_id.in_(
                sa.bindparam('network_ids', expanding=True)))
        db_objs = query(session).params(
            network_ids=network_ids).all()

        query = BAKERY(lambda s: s.query(
            NetworkExtensionCidrDb))
        query += lambda q: q.filter(
            NetworkExtensionCidrDb.network_id.in_(
                sa.bindparam('network_ids', expanding=True)))
        db_cidrs = query(session).params(
            network_ids=network_ids).all()

        query = BAKERY(lambda s: s.query(
            NetworkExtNestedDomainAllowedVlansDb))
        query += lambda q: q.filter(
            NetworkExtNestedDomainAllowedVlansDb.network_id.in_(
                sa.bindparam('network_ids', expanding=True)))
        db_vlans = query(session).params(
            network_ids=network_ids).all()

        query = BAKERY(lambda s: s.query(
            NetworkExtExtraContractDb))
        query += lambda q: q.filter(
            NetworkExtExtraContractDb.network_id.in_(
                sa.bindparam('network_ids', expanding=True)))
        db_contracts = query(session).params(
            network_ids=network_ids).all()

        query = BAKERY(lambda s: s.query(
            NetworkExtEpgContractMasterDb))
        query += lambda q: q.filter(
            NetworkExtEpgContractMasterDb.network_id.in_(
                sa.bindparam('network_ids', expanding=True)))
        db_masters = query(session).params(
            network_ids=network_ids).all()

        query = BAKERY(lambda s: s.query(
            NetworkExtensionNoNatCidrsDb))
        query += lambda q: q.filter(
            NetworkExtensionNoNatCidrsDb.network_id.in_(
                sa.bindparam('network_ids', expanding=True)))
        db_no_nat_cidrs = query(session).params(
            network_ids=network_ids).all()

        cidrs_by_net_id = {}
        vlans_by_net_id = {}
        contracts_by_net_id = {}
        masters_by_net_id = {}
        no_nat_cidrs_by_net_id = {}
        for db_cidr in db_cidrs:
            cidrs_by_net_id.setdefault(db_cidr.network_id, []).append(
                db_cidr)
        for db_vlan in db_vlans:
            vlans_by_net_id.setdefault(db_vlan.network_id, []).append(
                db_vlan)
        for db_contract in db_contracts:
            contracts_by_net_id.setdefault(db_contract.network_id, []).append(
                db_contract)
        for db_master in db_masters:
            masters_by_net_id.setdefault(db_master.network_id, []).append(
                db_master)
        for db_no_nat_cidr in db_no_nat_cidrs:
            no_nat_cidrs_by_net_id.setdefault(db_no_nat_cidr.network_id,
            []).append(db_no_nat_cidr)

        result = {}
        for db_obj in db_objs:
            net_id = db_obj.network_id
            result.setdefault(net_id, self.make_network_extn_db_conf_dict(
                db_obj, cidrs_by_net_id.get(net_id, []),
                vlans_by_net_id.get(net_id, []),
                contracts_by_net_id.get(net_id, []),
                masters_by_net_id.get(net_id, []),
                no_nat_cidrs_by_net_id.get(net_id, [])))
        return result

    def make_network_extn_db_conf_dict(self, ext_db, db_cidrs, db_vlans,
                                       db_contracts, db_masters,
                                       db_no_nat_cidrs):
        net_res = {}
        db_obj = ext_db
        if db_obj:
            self._set_if_not_none(net_res, cisco_apic.EXTERNAL_NETWORK,
                                  db_obj['external_network_dn'])
            self._set_if_not_none(net_res, cisco_apic.BD,
                                  db_obj['bridge_domain_dn'])
            self._set_if_not_none(net_res, cisco_apic.NAT_TYPE,
                                  db_obj['nat_type'])
            self._set_if_not_none(net_res, cisco_apic.SVI, db_obj['svi'])
            net_res[cisco_apic.BGP] = db_obj['bgp_enable']
            net_res[cisco_apic.BGP_TYPE] = db_obj['bgp_type']
            net_res[cisco_apic.BGP_ASN] = db_obj['bgp_asn']
            net_res[cisco_apic.NESTED_DOMAIN_NAME] = (
                    db_obj['nested_domain_name'])
            net_res[cisco_apic.NESTED_DOMAIN_TYPE] = (
                    db_obj['nested_domain_type'])
            net_res[cisco_apic.NESTED_DOMAIN_INFRA_VLAN] = (
                    db_obj['nested_domain_infra_vlan'])
            net_res[cisco_apic.NESTED_DOMAIN_SERVICE_VLAN] = (
                    db_obj['nested_domain_service_vlan'])
            net_res[cisco_apic.NESTED_DOMAIN_NODE_NETWORK_VLAN] = (
                    db_obj['nested_domain_node_network_vlan'])
            net_res[cisco_apic.NESTED_DOMAIN_ALLOWED_VLANS] = [
                c.vlan for c in db_vlans]
            net_res[cisco_apic.EXTRA_PROVIDED_CONTRACTS] = [
                c.contract_name for c in db_contracts if c.provides]
            net_res[cisco_apic.EXTRA_CONSUMED_CONTRACTS] = [
                c.contract_name for c in db_contracts if not c.provides]
            net_res[cisco_apic.EPG_CONTRACT_MASTERS] = [
                {'app_profile_name': m.app_profile_name,
                 'name': m.name} for m in db_masters]
            net_res[cisco_apic.POLICY_ENFORCEMENT_PREF] = db_obj[
                'policy_enforcement_pref']
            net_res[cisco_apic.NO_NAT_CIDRS] = [
                c.cidr for c in db_no_nat_cidrs]
            net_res[cisco_apic.MULTI_EXT_NETS] = db_obj['multi_ext_nets']
        if net_res.get(cisco_apic.EXTERNAL_NETWORK):
            net_res[cisco_apic.EXTERNAL_CIDRS] = [c.cidr for c in db_cidrs]
        return net_res

    def set_network_extn_db(self, session, network_id, res_dict):
        with session.begin(subtransactions=True):
            query = BAKERY(lambda s: s.query(
                NetworkExtensionDb))
            query += lambda q: q.filter_by(
                network_id=sa.bindparam('network_id'))
            db_obj = query(session).params(
                network_id=network_id).first()

            db_obj = db_obj or NetworkExtensionDb(network_id=network_id)
            if cisco_apic.EXTERNAL_NETWORK in res_dict:
                db_obj['external_network_dn'] = (
                    res_dict[cisco_apic.EXTERNAL_NETWORK])
            if cisco_apic.BD in res_dict:
                db_obj['bridge_domain_dn'] = (
                    res_dict[cisco_apic.BD])
            if cisco_apic.NAT_TYPE in res_dict:
                db_obj['nat_type'] = res_dict[cisco_apic.NAT_TYPE]
            if cisco_apic.SVI in res_dict:
                db_obj['svi'] = res_dict[cisco_apic.SVI]
            if cisco_apic.BGP in res_dict:
                db_obj['bgp_enable'] = res_dict[cisco_apic.BGP]
            if cisco_apic.BGP_TYPE in res_dict:
                db_obj['bgp_type'] = res_dict[cisco_apic.BGP_TYPE]
            if cisco_apic.BGP_ASN in res_dict:
                db_obj['bgp_asn'] = res_dict[cisco_apic.BGP_ASN]
            if cisco_apic.NESTED_DOMAIN_NAME in res_dict:
                db_obj['nested_domain_name'] = res_dict[
                        cisco_apic.NESTED_DOMAIN_NAME]
            if cisco_apic.NESTED_DOMAIN_TYPE in res_dict:
                db_obj['nested_domain_type'] = res_dict[
                        cisco_apic.NESTED_DOMAIN_TYPE]
            if cisco_apic.NESTED_DOMAIN_INFRA_VLAN in res_dict:
                db_obj['nested_domain_infra_vlan'] = res_dict[
                        cisco_apic.NESTED_DOMAIN_INFRA_VLAN]
            if cisco_apic.NESTED_DOMAIN_SERVICE_VLAN in res_dict:
                db_obj['nested_domain_service_vlan'] = res_dict[
                        cisco_apic.NESTED_DOMAIN_SERVICE_VLAN]
            if cisco_apic.NESTED_DOMAIN_NODE_NETWORK_VLAN in res_dict:
                db_obj['nested_domain_node_network_vlan'] = res_dict[
                        cisco_apic.NESTED_DOMAIN_NODE_NETWORK_VLAN]
            if cisco_apic.POLICY_ENFORCEMENT_PREF in res_dict:
                db_obj['policy_enforcement_pref'] = res_dict[
                        cisco_apic.POLICY_ENFORCEMENT_PREF]
            if cisco_apic.MULTI_EXT_NETS in res_dict:
                db_obj['multi_ext_nets'] = res_dict[cisco_apic.MULTI_EXT_NETS]
            session.add(db_obj)

            if cisco_apic.EXTERNAL_CIDRS in res_dict:
                self._update_list_attr(session, NetworkExtensionCidrDb, 'cidr',
                                       res_dict[cisco_apic.EXTERNAL_CIDRS],
                                       network_id=network_id)

            if cisco_apic.NESTED_DOMAIN_ALLOWED_VLANS in res_dict:
                self._update_list_attr(
                        session, NetworkExtNestedDomainAllowedVlansDb, 'vlan',
                        res_dict[cisco_apic.NESTED_DOMAIN_ALLOWED_VLANS],
                        network_id=network_id)

            if cisco_apic.EXTRA_PROVIDED_CONTRACTS in res_dict:
                self._update_list_attr(
                        session, NetworkExtExtraContractDb, 'contract_name',
                        res_dict[cisco_apic.EXTRA_PROVIDED_CONTRACTS],
                        network_id=network_id, provides=True)

            if cisco_apic.EXTRA_CONSUMED_CONTRACTS in res_dict:
                self._update_list_attr(
                        session, NetworkExtExtraContractDb, 'contract_name',
                        res_dict[cisco_apic.EXTRA_CONSUMED_CONTRACTS],
                        network_id=network_id, provides=False)

            if cisco_apic.EPG_CONTRACT_MASTERS in res_dict:
                self._update_dict_attr(
                        session, NetworkExtEpgContractMasterDb,
                        ('app_profile_name', 'name'),
                        res_dict[cisco_apic.EPG_CONTRACT_MASTERS],
                        network_id=network_id)

            if cisco_apic.NO_NAT_CIDRS in res_dict:
                self._update_list_attr(session, NetworkExtensionNoNatCidrsDb,
                                       'cidr',
                                       res_dict[cisco_apic.NO_NAT_CIDRS],
                                       network_id=network_id)

    def get_network_ids_by_ext_net_dn(self, session, dn, lock_update=False):
        query = BAKERY(lambda s: s.query(
            NetworkExtensionDb.network_id))
        query += lambda q: q.filter_by(
            external_network_dn=sa.bindparam('dn'))
        if lock_update:
            # REVISIT: Eliminate locking.
            query += lambda q: q.with_for_update()
        ids = query(session).params(dn=dn)

        return [i[0] for i in ids]

    def get_network_ids_by_l3out_dn(self, session, dn, lock_update=False):
        query = BAKERY(lambda s: s.query(
            NetworkExtensionDb.network_id))
        query += lambda q: q.filter(
            NetworkExtensionDb.external_network_dn.like(
                sa.bindparam('dn') + "/%"))
        if lock_update:
            # REVISIT: Eliminate locking.
            query += lambda q: q.with_for_update()
        ids = query(session).params(dn=dn)

        return [i[0] for i in ids]

    def get_svi_network_ids_by_l3out_dn(self, session, dn, lock_update=False):
        query = BAKERY(lambda s: s.query(
            NetworkExtensionDb.network_id))
        query += lambda q: q.filter(
            NetworkExtensionDb.external_network_dn.like(
                sa.bindparam('dn') + "/%"),
            NetworkExtensionDb.svi == true())
        if lock_update:
            # REVISIT: Eliminate locking.
            query += lambda q: q.with_for_update()
        ids = query(session).params(dn=dn)

        return [i[0] for i in ids]

    def get_external_cidrs_by_ext_net_dn(self, session, dn, lock_update=False):
        ctab = NetworkExtensionCidrDb
        ntab = NetworkExtensionDb

        query = BAKERY(lambda s: s.query(
            ctab.cidr))
        query += lambda q: q.join(
            ntab,
            ntab.network_id == ctab.network_id)
        query += lambda q: q.filter(
            ntab.external_network_dn == sa.bindparam('dn'))
        query += lambda q: q.distinct()
        if lock_update:
            # REVISIT: Eliminate locking.
            query += lambda q: q.with_for_update()
        cidrs = query(session).params(dn=dn)

        return [c[0] for c in cidrs]

    def get_external_cidrs_by_net_id(self, session, nid):
        query = BAKERY(lambda s: s.query(
            NetworkExtensionCidrDb.cidr))
        query += lambda q: q.filter_by(
            network_id=sa.bindparam('nid'))
        cidrs = query(session).params(nid=nid)

        return [i[0] for i in cidrs]

    def get_subnet_extn_db(self, session, subnet_id):
        query = BAKERY(lambda s: s.query(
            SubnetExtensionDb))
        query += lambda q: q.filter_by(
            subnet_id=sa.bindparam('subnet_id'))
        db_obj = query(session).params(
            subnet_id=subnet_id).first()

        result = {}
        if db_obj:
            self._set_if_not_none(result, cisco_apic.SNAT_HOST_POOL,
                                  db_obj['snat_host_pool'])
            self._set_if_not_none(result, cisco_apic.ACTIVE_ACTIVE_AAP,
                                  db_obj['active_active_aap'])
            self._set_if_not_none(result, cisco_apic.SNAT_SUBNET_ONLY,
                                  db_obj['snat_subnet_only'])
            self._set_if_not_none(result, cisco_apic.EPG_SUBNET,
                                  db_obj['epg_subnet'])
        return result

    def set_subnet_extn_db(self, session, subnet_id, res_dict):
        query = BAKERY(lambda s: s.query(
            SubnetExtensionDb))
        query += lambda q: q.filter_by(
            subnet_id=sa.bindparam('subnet_id'))
        db_obj = query(session).params(
            subnet_id=subnet_id).first()

        db_obj = db_obj or SubnetExtensionDb(subnet_id=subnet_id)
        if cisco_apic.SNAT_HOST_POOL in res_dict:
            db_obj['snat_host_pool'] = res_dict[cisco_apic.SNAT_HOST_POOL]
        if cisco_apic.ACTIVE_ACTIVE_AAP in res_dict:
            db_obj['active_active_aap'] = res_dict[
                                            cisco_apic.ACTIVE_ACTIVE_AAP]
        if cisco_apic.SNAT_SUBNET_ONLY in res_dict:
            db_obj['snat_subnet_only'] = res_dict[
                                            cisco_apic.SNAT_SUBNET_ONLY]
        if cisco_apic.EPG_SUBNET in res_dict:
            db_obj['epg_subnet'] = res_dict[cisco_apic.EPG_SUBNET]
        session.add(db_obj)

    def get_router_extn_db(self, session, router_id):
        query = BAKERY(lambda s: s.query(
            RouterExtensionContractDb))
        query += lambda q: q.filter_by(
            router_id=sa.bindparam('router_id'))
        db_contracts = query(session).params(
            router_id=router_id).all()

        return {cisco_apic_l3.EXTERNAL_PROVIDED_CONTRACTS:
                [c['contract_name'] for c in db_contracts if c['provides']],
                cisco_apic_l3.EXTERNAL_CONSUMED_CONTRACTS:
                [c['contract_name'] for c in db_contracts
                 if not c['provides']]}

    def get_router_extn_db_bulk(self, session, router_ids):
        query = BAKERY(lambda s: s.query(
            RouterExtensionContractDb))
        query += lambda q: q.filter(
            RouterExtensionContractDb.router_id.in_(
                sa.bindparam('router_ids', expanding=True)))
        db_contracts = query(session).params(
            router_ids=router_ids).all()

        attr_dict = defaultdict(dict)
        for db_contract in db_contracts:
            router_id = db_contract['router_id']
            p_contracts = attr_dict[router_id].setdefault(
                cisco_apic_l3.EXTERNAL_PROVIDED_CONTRACTS, [])
            c_contracts = attr_dict[router_id].setdefault(
                cisco_apic_l3.EXTERNAL_CONSUMED_CONTRACTS, [])
            if db_contract['provides']:
                p_contracts.append(db_contract['contract_name'])
            else:
                c_contracts.append(db_contract['contract_name'])
        return attr_dict

    def get_network_ids_and_multi_by_l3out_dn(self, session, dn):
        query = BAKERY(lambda s: s.query(
            NetworkExtensionDb.network_id,
            NetworkExtensionDb.multi_ext_nets))
        query += lambda q: q.filter(
            NetworkExtensionDb.external_network_dn.like(
                sa.bindparam('dn') + "/%"))
        ids_and_multis = query(session).params(dn=dn)

        return [(i[0], i[1]) for i in ids_and_multis]

    def _update_list_attr(self, session, db_model, column,
                          new_values, **filters):
        if new_values is None:
            return

        # REVISIT: Can this query be baked?
        rows = session.query(db_model).filter_by(**filters).all()

        new_values = set(new_values)
        for r in rows:
            if r[column] in new_values:
                new_values.discard(r[column])
            else:
                session.delete(r)
        for v in new_values:
            attr = {column: v}
            attr.update(filters)
            db_obj = db_model(**attr)
            session.add(db_obj)

    def _update_dict_attr(self, session, db_model, keys,
                          new_values, **filters):
        if new_values is None:
            return

        # REVISIT: Can this query be baked?
        rows = session.query(db_model).filter_by(**filters).all()

        # remove duplicates, may change order
        new_values = [dict(t) for t in {tuple(d.items()) for d in new_values}]
        # Updates are deletions with additions, so to ensure that
        # the delete happens before a subsequent addtion, we create
        # a subtransaction
        with session.begin(subtransactions=True):
            for r in rows:
                curr_obj = {key: r[key] for key in keys}
                if curr_obj in new_values:
                    new_values.remove(curr_obj)
                else:
                    session.delete(r)
        for v in new_values:
            v.update(filters)
            db_obj = db_model(**v)
            session.add(db_obj)

    def set_router_extn_db(self, session, router_id, res_dict):
        with session.begin(subtransactions=True):
            if cisco_apic_l3.EXTERNAL_PROVIDED_CONTRACTS in res_dict:
                self._update_list_attr(session, RouterExtensionContractDb,
                   'contract_name',
                   res_dict[cisco_apic_l3.EXTERNAL_PROVIDED_CONTRACTS],
                   router_id=router_id, provides=True)
            if cisco_apic_l3.EXTERNAL_CONSUMED_CONTRACTS in res_dict:
                self._update_list_attr(session, RouterExtensionContractDb,
                    'contract_name',
                    res_dict[cisco_apic_l3.EXTERNAL_CONSUMED_CONTRACTS],
                    router_id=router_id, provides=False)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""change common ap name

Revision ID: facc1ababba1
Revises: d978a7a73785
Create Date: 2017-05-15 00:00:00.000000

"""

# revision identifiers, used by Alembic.
revision = 'facc1ababba1'
down_revision = 'd978a7a73785'

from alembic import op
import sqlalchemy as sa


def upgrade():
    # See if AIM is being used, and if so, migrate data.
    bind = op.get_bind()
    insp = sa.engine.reflection.Inspector.from_engine(bind)
    if 'aim_tenants' in insp.get_table_names():
        # Note - this cannot be imported unless we know the
        # apic_aim mechanism driver is deployed, since the AIM
        # library may not be installed.
        from gbpservice.neutron.plugins.ml2plus.drivers.apic_aim import (
            data_migrations)

        session = sa.orm.Session(bind=bind, autocommit=True)
        data_migrations.do_ap_name_change(session)


def downgrade():
    pass



# Copyright 2014 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""gbp_shared_attribute

Revision ID: f4d890a9c126
Revises: d595542cf3f5
Create Date: 2014-11-12 21:13:08.98888

"""

# revision identifiers, used by Alembic.
revision = 'f4d890a9c126'
down_revision = 'd595542cf3f5'


from alembic import op
import sqlalchemy as sa


def upgrade():

    op.add_column(
        'gp_policy_target_groups',
        sa.Column('shared', sa.Boolean)
    )

    op.add_column(
        'gp_l2_policies',
        sa.Column('shared', sa.Boolean)
    )

    op.add_column(
        'gp_l3_policies',
        sa.Column('shared', sa.Boolean)
    )

    op.add_column(
        'gp_policy_rule_sets',
        sa.Column('shared', sa.Boolean)
    )

    op.add_column(
        'gp_policy_rules',
        sa.Column('shared', sa.Boolean)
    )

    op.add_column(
        'gp_policy_classifiers',
        sa.Column('shared', sa.Boolean)
    )

    op.add_column(
        'gp_policy_actions',
        sa.Column('shared', sa.Boolean)
    )

    op.add_column(
        'gp_network_service_policies',
        sa.Column('shared', sa.Boolean)
    )


def downgrade():
    op.drop_column('gp_network_service_policies', 'shared')
    op.drop_column('gp_policy_actions', 'shared')
    op.drop_column('gp_policy_classifiers', 'shared')
    op.drop_column('gp_policy_rules', 'shared')
    op.drop_column('gp_policy_rule_sets', 'shared')
    op.drop_column('gp_l3_policies', 'shared')
    op.drop_column('gp_l2_policies', 'shared')
    op.drop_column('gp_policy_target_groups', 'shared')
